#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import time
import unittest
from typing import Optional, Tuple, Union
from unittest.mock import MagicMock, call

import evdev
import gi
from evdev.ecodes import KEY_A, KEY_B, KEY_C

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GLib", "2.0")
gi.require_version("GtkSource", "4")
from gi.repository import Gtk, GLib, GtkSource, Gdk

from tests.lib.spy import spy
from tests.lib.logger import logger

from inputremapper.gui.controller import Controller
from inputremapper.configs.keyboard_layout import XKB_KEYCODE_OFFSET
from inputremapper.gui.utils import CTX_ERROR, CTX_WARNING, gtk_iteration
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import (
    UInputsData,
    GroupsData,
    GroupData,
    PresetData,
    StatusData,
    CombinationUpdate,
    DoStackSwitch,
)
from inputremapper.groups import DeviceType
from inputremapper.gui.components.editor import (
    TargetSelection,
    MappingListBox,
    MappingSelectionLabel,
    CodeEditor,
    RecordingToggle,
    AutoloadSwitch,
    ReleaseCombinationSwitch,
    CombinationListbox,
    InputConfigEntry,
    AnalogInputSwitch,
    TriggerThresholdInput,
    ReleaseTimeoutInput,
    OutputAxisSelector,
    KeyAxisStackSwitcher,
    Sliders,
    TransformationDrawArea,
    RelativeInputCutoffInput,
    RecordingStatus,
    RequireActiveMapping,
    GdkEventRecorder,
)
from inputremapper.gui.components.main import Stack, StatusBar
from inputremapper.gui.components.common import FlowBoxEntry, Breadcrumbs
from inputremapper.gui.components.presets import PresetSelection
from inputremapper.gui.components.device_groups import (
    DeviceGroupEntry,
    DeviceGroupSelection,
)
from inputremapper.configs.mapping import MappingData
from inputremapper.configs.input_config import InputCombination, InputConfig
from tests.lib.test_setup import test_setup


class ComponentBaseTest(unittest.TestCase):
    """Test a gui component."""

    def setUp(self) -> None:
        self.message_broker = MessageBroker()
        self.controller_mock: Controller = MagicMock()

    def destroy_all_member_widgets(self):
        # destroy all Gtk Widgets that are stored in self
        # TODO why is this necessary?
        for attribute in dir(self):
            stuff = getattr(self, attribute, None)
            if isinstance(stuff, Gtk.Widget):
                logger.info('destroying member "%s" %s', attribute, stuff)
                GLib.timeout_add(0, stuff.destroy)
                setattr(self, attribute, None)

    def tearDown(self) -> None:
        super().tearDown()
        self.message_broker.signal(MessageType.terminate)

        # Shut down the gui properly
        self.destroy_all_member_widgets()
        GLib.timeout_add(0, Gtk.main_quit)

        # Gtk.main() will start the Gtk event loop and process all pending events.
        # So the gui will do whatever is queued up this ensures that the next tests
        # starts without pending events.
        Gtk.main()


class FlowBoxTestUtils:
    """Methods to test the FlowBoxes that contain presets and devices.

    Those are only used in tests, so I moved them here instead.
    """

    @staticmethod
    def set_active(flow_box: Gtk.FlowBox, name: str):
        """Change the currently selected group."""
        for child in flow_box.get_children():
            flow_box_entry: FlowBoxEntry = child.get_children()[0]
            flow_box_entry.set_active(flow_box_entry.name == name)

    @staticmethod
    def get_active_entry(flow_box: Gtk.FlowBox) -> Union[DeviceGroupEntry, None]:
        """Find the currently selected DeviceGroupEntry."""
        children = flow_box.get_children()

        if len(children) == 0:
            return None

        for child in children:
            flow_box_entry: FlowBoxEntry = child.get_children()[0]

            if flow_box_entry.get_active():
                return flow_box_entry

        raise AssertionError("Expected one entry to be selected.")

    @staticmethod
    def get_child_names(flow_box: Gtk.FlowBox):
        names = []
        for child in flow_box.get_children():
            flow_box_entry: FlowBoxEntry = child.get_children()[0]
            names.append(flow_box_entry.name)

        return names

    @staticmethod
    def get_child_icons(flow_box: Gtk.FlowBox):
        icon_names = []
        for child in flow_box.get_children():
            flow_box_entry: FlowBoxEntry = child.get_children()[0]
            icon_names.append(flow_box_entry.icon_name)

        return icon_names


@test_setup
class TestDeviceGroupSelection(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.FlowBox()
        self.selection = DeviceGroupSelection(
            self.message_broker,
            self.controller_mock,
            self.gui,
        )
        self.message_broker.publish(
            GroupsData(
                {
                    "foo": [DeviceType.GAMEPAD, DeviceType.KEYBOARD],
                    "bar": [],
                    "baz": [DeviceType.GRAPHICS_TABLET],
                }
            )
        )

    def get_displayed_group_keys_and_icons(self):
        """Get a list of all group_keys and icons of the displayed groups."""
        group_keys = []
        icons = []
        for child in self.gui.get_children():
            device_group_entry = child.get_children()[0]
            group_keys.append(device_group_entry.group_key)
            icons.append(device_group_entry.icon_name)

        return group_keys, icons

    def test_populates_devices(self):
        # tests that all devices sent via the broker end up in the gui
        group_keys, icons = self.get_displayed_group_keys_and_icons()
        self.assertEqual(group_keys, ["foo", "bar", "baz"])
        self.assertEqual(icons, ["input-gaming", None, "input-tablet"])

        self.message_broker.publish(
            GroupsData(
                {
                    "kuu": [DeviceType.KEYBOARD],
                    "qux": [DeviceType.GAMEPAD],
                }
            )
        )

        group_keys, icons = self.get_displayed_group_keys_and_icons()
        self.assertEqual(group_keys, ["kuu", "qux"])
        self.assertEqual(icons, ["input-keyboard", "input-gaming"])

    def test_selects_correct_device(self):
        self.message_broker.publish(GroupData("bar", ()))
        self.assertEqual(FlowBoxTestUtils.get_active_entry(self.gui).group_key, "bar")
        self.message_broker.publish(GroupData("baz", ()))
        self.assertEqual(FlowBoxTestUtils.get_active_entry(self.gui).group_key, "baz")

    def test_loads_group(self):
        FlowBoxTestUtils.set_active(self.gui, "bar")
        self.controller_mock.load_group.assert_called_once_with("bar")

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(GroupData("bar", ()))
        self.controller_mock.load_group.assert_not_called()


@test_setup
class TestTargetSelection(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.ComboBox()
        self.selection = TargetSelection(
            self.message_broker, self.controller_mock, self.gui
        )
        self.message_broker.publish(
            UInputsData(
                {
                    "foo": {},
                    "bar": {},
                    "baz": {},
                }
            )
        )

    def test_populates_devices(self):
        names = [row[0] for row in self.gui.get_model()]
        self.assertEqual(names, ["foo", "bar", "baz"])

        self.message_broker.publish(
            UInputsData(
                {
                    "kuu": {},
                    "qux": {},
                }
            )
        )
        names = [row[0] for row in self.gui.get_model()]
        self.assertEqual(names, ["kuu", "qux"])

    def test_updates_mapping(self):
        self.gui.set_active_id("baz")
        self.controller_mock.update_mapping.assert_called_once_with(target_uinput="baz")

    def test_selects_correct_target(self):
        self.message_broker.publish(MappingData(target_uinput="baz"))
        self.assertEqual(self.gui.get_active_id(), "baz")
        self.message_broker.publish(MappingData(target_uinput="bar"))
        self.assertEqual(self.gui.get_active_id(), "bar")

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(MappingData(target_uinput="baz"))
        self.controller_mock.update_mapping.assert_not_called()


@test_setup
class TestPresetSelection(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.FlowBox()
        self.selection = PresetSelection(
            self.message_broker, self.controller_mock, self.gui
        )
        self.message_broker.publish(GroupData("foo", ("preset1", "preset2")))

    def test_populates_presets(self):
        names = FlowBoxTestUtils.get_child_names(self.gui)
        self.assertEqual(names, ["preset1", "preset2"])

        self.message_broker.publish(GroupData("foo", ("preset3", "preset4")))
        names = FlowBoxTestUtils.get_child_names(self.gui)
        self.assertEqual(names, ["preset3", "preset4"])

    def test_selects_preset(self):
        self.message_broker.publish(
            PresetData(
                "preset2",
                (
                    MappingData(
                        name="m1",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=2)]
                        ),
                    ),
                ),
            )
        )
        self.assertEqual(FlowBoxTestUtils.get_active_entry(self.gui).name, "preset2")

        self.message_broker.publish(
            PresetData(
                "preset1",
                (
                    MappingData(
                        name="m1",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=2)]
                        ),
                    ),
                ),
            )
        )
        self.assertEqual(FlowBoxTestUtils.get_active_entry(self.gui).name, "preset1")

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(
            PresetData(
                "preset2",
                (
                    MappingData(
                        name="m1",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=2)]
                        ),
                    ),
                ),
            )
        )
        self.controller_mock.load_preset.assert_not_called()

    def test_loads_preset(self):
        FlowBoxTestUtils.set_active(self.gui, "preset2")
        self.controller_mock.load_preset.assert_called_once_with("preset2")


@test_setup
class TestMappingListbox(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.ListBox()
        self.listbox = MappingListBox(
            self.message_broker, self.controller_mock, self.gui
        )

        self.message_broker.publish(
            PresetData(
                "preset1",
                (
                    MappingData(
                        name="mapping1",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_C)]
                        ),
                    ),
                    MappingData(
                        name="",
                        input_combination=InputCombination(
                            [
                                InputConfig(type=1, code=KEY_A),
                                InputConfig(type=1, code=KEY_B),
                            ]
                        ),
                    ),
                    MappingData(
                        name="mapping2",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_B)]
                        ),
                    ),
                ),
            )
        )

    def get_selected_row(self) -> MappingSelectionLabel:
        for label in self.gui.get_children():
            if label.is_selected():
                return label

        raise Exception("Expected one MappingSelectionLabel to be selected")

    def select_row(self, combination: InputCombination):
        def select(label_: MappingSelectionLabel):
            if label_.combination == combination:
                self.gui.select_row(label_)

        for label in self.gui.get_children():
            select(label)

    def test_populates_listbox(self):
        labels = {row.name for row in self.gui.get_children()}
        self.assertEqual(labels, {"mapping1", "mapping2", "a + b"})

    def test_alphanumerically_sorted(self):
        labels = [row.name for row in self.gui.get_children()]
        self.assertEqual(labels, ["a + b", "mapping1", "mapping2"])

    def test_activates_correct_row(self):
        self.message_broker.publish(
            MappingData(
                name="mapping1",
                input_combination=InputCombination([InputConfig(type=1, code=KEY_C)]),
            )
        )
        selected = self.get_selected_row()
        self.assertEqual(selected.name, "mapping1")
        self.assertEqual(
            selected.combination,
            InputCombination([InputConfig(type=1, code=KEY_C)]),
        )

    def test_loads_mapping(self):
        self.select_row(InputCombination([InputConfig(type=1, code=KEY_B)]))
        self.controller_mock.load_mapping.assert_called_once_with(
            InputCombination([InputConfig(type=1, code=KEY_B)])
        )

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(
            MappingData(
                name="mapping1",
                input_combination=InputCombination([InputConfig(type=1, code=KEY_C)]),
            )
        )
        self.controller_mock.load_mapping.assert_not_called()

    def test_sorts_empty_mapping_to_bottom(self):
        self.message_broker.publish(
            PresetData(
                "preset1",
                (
                    MappingData(
                        name="qux",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_C)]
                        ),
                    ),
                    MappingData(
                        name="foo",
                        input_combination=InputCombination.empty_combination(),
                    ),
                    MappingData(
                        name="bar",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_B)]
                        ),
                    ),
                ),
            )
        )
        bottom_row: MappingSelectionLabel = self.gui.get_row_at_index(2)
        self.assertEqual(bottom_row.combination, InputCombination.empty_combination())
        self.message_broker.publish(
            PresetData(
                "preset1",
                (
                    MappingData(
                        name="foo",
                        input_combination=InputCombination.empty_combination(),
                    ),
                    MappingData(
                        name="qux",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_C)]
                        ),
                    ),
                    MappingData(
                        name="bar",
                        input_combination=InputCombination(
                            [InputConfig(type=1, code=KEY_B)]
                        ),
                    ),
                ),
            )
        )
        bottom_row: MappingSelectionLabel = self.gui.get_row_at_index(2)
        self.assertEqual(bottom_row.combination, InputCombination.empty_combination())


@test_setup
class TestMappingSelectionLabel(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.ListBox()
        self.mapping_selection_label = MappingSelectionLabel(
            self.message_broker,
            self.controller_mock,
            "",
            InputCombination(
                [
                    InputConfig(type=1, code=KEY_A),
                    InputConfig(type=1, code=KEY_B),
                ]
            ),
        )
        self.gui.insert(self.mapping_selection_label, -1)

    def assert_edit_mode(self):
        self.assertTrue(self.mapping_selection_label.name_input.get_visible())
        self.assertFalse(self.mapping_selection_label.label.get_visible())

    def assert_selected(self):
        self.assertTrue(self.mapping_selection_label.label.get_visible())
        self.assertFalse(self.mapping_selection_label.name_input.get_visible())

    def test_repr(self):
        self.mapping_selection_label.name = "name"
        self.assertIn("name", repr(self.mapping_selection_label))
        self.assertIn("KEY_A", repr(self.mapping_selection_label))
        self.assertIn("KEY_B", repr(self.mapping_selection_label))

    def test_shows_combination_without_name(self):
        self.assertEqual(self.mapping_selection_label.label.get_label(), "a + b")

    def test_shows_name_when_given(self):
        self.gui = MappingSelectionLabel(
            self.message_broker,
            self.controller_mock,
            "foo",
            InputCombination(
                (
                    InputConfig(type=1, code=KEY_A),
                    InputConfig(type=1, code=KEY_B),
                )
            ),
        )
        self.assertEqual(self.gui.label.get_label(), "foo")

    def test_updates_combination_when_selected(self):
        self.gui.select_row(self.mapping_selection_label)
        self.assertEqual(
            self.mapping_selection_label.combination,
            InputCombination(
                (
                    InputConfig(type=1, code=KEY_A),
                    InputConfig(type=1, code=KEY_B),
                )
            ),
        )
        self.message_broker.publish(
            CombinationUpdate(
                InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
                InputCombination([InputConfig(type=1, code=KEY_A)]),
            )
        )
        self.assertEqual(
            self.mapping_selection_label.combination,
            InputCombination([InputConfig(type=1, code=KEY_A)]),
        )

    def test_doesnt_update_combination_when_not_selected(self):
        self.assertEqual(
            self.mapping_selection_label.combination,
            InputCombination(
                (
                    InputConfig(type=1, code=KEY_A),
                    InputConfig(type=1, code=KEY_B),
                )
            ),
        )
        self.message_broker.publish(
            CombinationUpdate(
                InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
                InputCombination([InputConfig(type=1, code=KEY_A)]),
            )
        )
        self.assertEqual(
            self.mapping_selection_label.combination,
            InputCombination(
                (
                    InputConfig(type=1, code=KEY_A),
                    InputConfig(type=1, code=KEY_B),
                )
            ),
        )

    def test_updates_name_when_mapping_changed_and_combination_matches(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
                name="foo",
            )
        )
        self.assertEqual(self.mapping_selection_label.label.get_label(), "foo")

    def test_ignores_mapping_when_combination_does_not_match(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_C),
                    )
                ),
                name="foo",
            )
        )
        self.assertEqual(self.mapping_selection_label.label.get_label(), "a + b")

    def test_edit_button_visibility(self):
        # start off invisible
        self.assertFalse(self.mapping_selection_label.edit_btn.get_visible())

        # load the mapping associated with the ListBoxRow
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.assertTrue(self.mapping_selection_label.edit_btn.get_visible())

        # load a different row
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_C),
                    )
                ),
            )
        )
        self.assertFalse(self.mapping_selection_label.edit_btn.get_visible())

    def test_enter_edit_mode_focuses_name_input(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.mapping_selection_label.edit_btn.clicked()
        self.controller_mock.set_focus.assert_called_once_with(
            self.mapping_selection_label.name_input
        )

    def test_enter_edit_mode_updates_visibility(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.assert_selected()
        self.mapping_selection_label.edit_btn.clicked()
        self.assert_edit_mode()
        self.mapping_selection_label.name_input.activate()  # aka hit the return key
        self.assert_selected()

    def test_leaves_edit_mode_on_esc(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.mapping_selection_label.edit_btn.clicked()
        self.assert_edit_mode()
        self.mapping_selection_label.name_input.set_text("foo")

        event = Gdk.Event()
        event.key.keyval = Gdk.KEY_Escape
        # send the "key-press-event"
        self.mapping_selection_label._on_gtk_rename_abort(None, event.key)
        self.assert_selected()
        self.assertEqual(self.mapping_selection_label.label.get_text(), "a + b")
        self.controller_mock.update_mapping.assert_not_called()

    def test_update_name(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.mapping_selection_label.edit_btn.clicked()

        self.mapping_selection_label.name_input.set_text("foo")
        self.mapping_selection_label.name_input.activate()
        self.controller_mock.update_mapping.assert_called_once_with(name="foo")

    def test_name_input_contains_combination_when_name_not_set(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
            )
        )
        self.mapping_selection_label.edit_btn.clicked()
        self.assertEqual(self.mapping_selection_label.name_input.get_text(), "a + b")

    def test_name_input_contains_name(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
                name="foo",
            )
        )
        self.mapping_selection_label.edit_btn.clicked()
        self.assertEqual(self.mapping_selection_label.name_input.get_text(), "foo")

    def test_removes_name_when_name_matches_combination(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=KEY_A),
                        InputConfig(type=1, code=KEY_B),
                    )
                ),
                name="foo",
            )
        )
        self.mapping_selection_label.edit_btn.clicked()
        self.mapping_selection_label.name_input.set_text("a + b")
        self.mapping_selection_label.name_input.activate()
        self.controller_mock.update_mapping.assert_called_once_with(name="")


@test_setup
class TestGdkEventRecorder(ComponentBaseTest):
    def _emit_key(self, window, code, type_):
        event = Gdk.Event()
        event.type = type_
        event.hardware_keycode = code + XKB_KEYCODE_OFFSET
        window.emit("event", event)
        gtk_iteration()

    def _emit_button(self, window, button, type_):
        event = Gdk.Event()
        event.type = type_
        event.button = button
        window.emit("event", event)
        gtk_iteration()

    def test_records_combinations(self):
        label = Gtk.Label()
        window = Gtk.Window()
        GdkEventRecorder(window, label)

        self._emit_key(window, KEY_A, Gdk.EventType.KEY_PRESS)
        self._emit_key(window, KEY_B, Gdk.EventType.KEY_PRESS)
        self.assertEqual(label.get_text(), "a + b")

        self._emit_key(window, KEY_A, Gdk.EventType.KEY_RELEASE)
        self._emit_key(window, KEY_B, Gdk.EventType.KEY_RELEASE)
        self.assertEqual(label.get_text(), "a + b")

        self._emit_key(window, KEY_C, Gdk.EventType.KEY_PRESS)
        self.assertEqual(label.get_text(), "c")

        # buttons
        self._emit_button(window, Gdk.BUTTON_PRIMARY, Gdk.EventType.BUTTON_PRESS)
        self._emit_button(window, Gdk.BUTTON_SECONDARY, Gdk.EventType.BUTTON_PRESS)
        self._emit_button(window, Gdk.BUTTON_MIDDLE, Gdk.EventType.BUTTON_PRESS)
        # no constants seem to exist, but this is the value that was observed during
        # usage:
        self._emit_button(window, 8, Gdk.EventType.BUTTON_PRESS)
        self._emit_button(window, 9, Gdk.EventType.BUTTON_PRESS)
        self.assertEqual(
            label.get_text(),
            "c + BTN_LEFT + BTN_RIGHT + BTN_MIDDLE + BTN_SIDE + BTN_EXTRA",
        )

        # releasing anything resets the combination
        self._emit_button(window, 9, Gdk.EventType.BUTTON_RELEASE)
        self._emit_key(window, KEY_A, Gdk.EventType.KEY_PRESS)
        self.assertEqual(label.get_text(), "a")


@test_setup
class TestCodeEditor(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = GtkSource.View()
        self.editor = CodeEditor(self.message_broker, self.controller_mock, self.gui)
        # TODO why is mocking this to False needed?
        self.controller_mock.is_empty_mapping.return_value = False

    def get_text(self) -> str:
        buffer = self.gui.get_buffer()
        return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)

    def test_shows_output_symbol(self):
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.assertEqual(self.get_text(), "foo")

    def test_shows_record_input_first_message_when_mapping_is_empty(self):
        self.controller_mock.is_empty_mapping.return_value = True
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.assertEqual(self.get_text(), "Record the input first")

    def test_active_when_mapping_is_not_empty(self):
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.assertTrue(self.gui.get_sensitive())
        self.assertEqual(self.gui.get_opacity(), 1)

    def test_expands_to_multiline(self):
        self.message_broker.publish(MappingData(output_symbol="foo\nbar"))
        self.assertIn("multiline", self.gui.get_style_context().list_classes())

    def test_shows_line_numbers_when_multiline(self):
        self.message_broker.publish(MappingData(output_symbol="foo\nbar"))
        self.assertTrue(self.gui.get_show_line_numbers())

    def test_no_multiline_when_macro_not_multiline(self):
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.assertNotIn("multiline", self.gui.get_style_context().list_classes())

    def test_no_line_numbers_macro_not_multiline(self):
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.assertFalse(self.gui.get_show_line_numbers())

    def test_shows_placeholder_when_mapping_has_no_output_symbol(self):
        self.message_broker.publish(MappingData())
        self.assertEqual(self.get_text(), self.editor.placeholder)

        # there are no side-effects because the placeholder is inserted:
        self.controller_mock.update_mapping.assert_not_called()

    def test_updates_mapping(self):
        self.message_broker.publish(MappingData())
        buffer = self.gui.get_buffer()
        self.controller_mock.update_mapping.assert_not_called()
        buffer.set_text("foo")
        call_args_list = self.controller_mock.update_mapping.call_args_list
        # this test emits 2 events for whatever reason, the first one with an empty
        # symbol. this doesn't actually seem to happen when using it.
        self.assertEqual(call_args_list[-1], call(output_symbol="foo"))

    def test_avoids_infinite_recursion_when_loading_mapping(self):
        self.message_broker.publish(MappingData(output_symbol="foo"))
        self.controller_mock.update_mapping.assert_not_called()

    def test_gets_focus_when_input_recording_finises(self):
        self.message_broker.signal(MessageType.recording_finished)
        self.controller_mock.set_focus.assert_called_once_with(self.gui)

    def test_placeholder(self):
        self.assertEqual(self.get_text(), self.editor.placeholder)

        window = Gtk.Window()
        window.add(self.gui)
        window.show_all()

        def focus():
            self.gui.grab_focus()
            # Do as many iterations as needed to make it work.
            gtk_iteration(15)

        def unfocus():
            window.set_focus(None)
            gtk_iteration(15)

        # clears the input when we enter the editor widget
        focus()
        self.assertEqual(self.get_text(), "")
        self.assertNotIn("opaque-text", self.gui.get_style_context().list_classes())

        # adds the placeholder back when we leave it
        unfocus()
        self.assertEqual(self.get_text(), self.editor.placeholder)
        self.assertIn("opaque-text", self.gui.get_style_context().list_classes())

        # if we enter text and then leave, it won't show the placeholder
        focus()
        self.assertEqual(self.get_text(), "")
        buffer = self.gui.get_buffer()
        buffer.set_text("foo")
        self.assertNotIn("opaque-text", self.gui.get_style_context().list_classes())
        unfocus()
        self.assertEqual(self.get_text(), "foo")
        self.assertNotIn("opaque-text", self.gui.get_style_context().list_classes())


@test_setup
class TestRecordingToggle(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()

        self.toggle_button = Gtk.ToggleButton()
        self.recording_toggle = RecordingToggle(
            self.message_broker,
            self.controller_mock,
            self.toggle_button,
        )

        self.label = Gtk.Label()
        self.recording_status = RecordingStatus(self.message_broker, self.label)

    def assert_not_recording(self):
        self.assertFalse(self.label.get_visible())
        self.assertFalse(self.toggle_button.get_active())

    def test_starts_recording(self):
        self.toggle_button.set_active(True)
        self.controller_mock.start_key_recording.assert_called_once()

    def test_stops_recording_when_clicked(self):
        self.toggle_button.set_active(True)
        self.toggle_button.set_active(False)
        self.controller_mock.stop_key_recording.assert_called_once()

    def test_not_recording_initially(self):
        self.assert_not_recording()

    def test_shows_recording_when_message_sent(self):
        self.assertFalse(self.label.get_visible())
        self.message_broker.signal(MessageType.recording_started)
        self.assertTrue(self.label.get_visible())

    def test_shows_not_recording_after_toggle(self):
        self.toggle_button.set_active(True)
        self.toggle_button.set_active(False)
        self.assert_not_recording()

    def test_shows_not_recording_when_recording_finished(self):
        self.toggle_button.set_active(True)
        self.message_broker.signal(MessageType.recording_finished)
        self.assert_not_recording()


@test_setup
class TestStatusBar(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Statusbar()
        self.err_icon = Gtk.Image()
        self.warn_icon = Gtk.Image()
        self.statusbar = StatusBar(
            self.message_broker,
            self.controller_mock,
            self.gui,
            self.err_icon,
            self.warn_icon,
        )
        self.message_broker.signal(MessageType.init)

    def assert_empty(self):
        self.assertFalse(self.err_icon.get_visible())
        self.assertFalse(self.warn_icon.get_visible())
        self.assertEqual(self.get_text(), "")
        self.assertIsNone(self.get_tooltip())

    def assert_error_status(self):
        self.assertTrue(self.err_icon.get_visible())
        self.assertFalse(self.warn_icon.get_visible())

    def assert_warning_status(self):
        self.assertFalse(self.err_icon.get_visible())
        self.assertTrue(self.warn_icon.get_visible())

    def get_text(self) -> str:
        return self.gui.get_message_area().get_children()[0].get_text()

    def get_tooltip(self) -> Optional[str]:
        return self.gui.get_tooltip_text()

    def test_starts_empty(self):
        self.assert_empty()

    def test_shows_error_status(self):
        self.message_broker.publish(StatusData(CTX_ERROR, "msg", "tooltip"))
        self.assertEqual(self.get_text(), "msg")
        self.assertEqual(self.get_tooltip(), "tooltip")
        self.assert_error_status()

    def test_shows_warning_status(self):
        self.message_broker.publish(StatusData(CTX_WARNING, "msg", "tooltip"))
        self.assertEqual(self.get_text(), "msg")
        self.assertEqual(self.get_tooltip(), "tooltip")
        self.assert_warning_status()

    def test_shows_newest_message(self):
        self.message_broker.publish(StatusData(CTX_ERROR, "msg", "tooltip"))
        self.message_broker.publish(StatusData(CTX_WARNING, "msg2", "tooltip2"))
        self.assertEqual(self.get_text(), "msg2")
        self.assertEqual(self.get_tooltip(), "tooltip2")
        self.assert_warning_status()

    def test_data_without_message_removes_messages(self):
        self.message_broker.publish(StatusData(CTX_WARNING, "msg", "tooltip"))
        self.message_broker.publish(StatusData(CTX_WARNING, "msg2", "tooltip2"))
        self.message_broker.publish(StatusData(CTX_WARNING))
        self.assert_empty()

    def test_restores_message_from_not_removed_ctx_id(self):
        self.message_broker.publish(StatusData(CTX_ERROR, "msg", "tooltip"))
        self.message_broker.publish(StatusData(CTX_WARNING, "msg2", "tooltip2"))
        self.message_broker.publish(StatusData(CTX_WARNING))
        self.assertEqual(self.get_text(), "msg")
        self.assert_error_status()

        # works also the other way round
        self.message_broker.publish(StatusData(CTX_ERROR))
        self.message_broker.publish(StatusData(CTX_WARNING, "msg", "tooltip"))
        self.message_broker.publish(StatusData(CTX_ERROR, "msg2", "tooltip2"))
        self.message_broker.publish(StatusData(CTX_ERROR))
        self.assertEqual(self.get_text(), "msg")
        self.assert_warning_status()

    def test_sets_msg_as_tooltip_if_tooltip_is_none(self):
        self.message_broker.publish(StatusData(CTX_ERROR, "msg"))
        self.assertEqual(self.get_tooltip(), "msg")


@test_setup
class TestAutoloadSwitch(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Switch()
        self.switch = AutoloadSwitch(
            self.message_broker, self.controller_mock, self.gui
        )

    def test_sets_autoload(self):
        self.gui.set_active(True)
        self.controller_mock.set_autoload.assert_called_once_with(True)
        self.controller_mock.reset_mock()
        self.gui.set_active(False)
        self.controller_mock.set_autoload.assert_called_once_with(False)

    def test_updates_state(self):
        self.message_broker.publish(PresetData(None, None, autoload=True))
        self.assertTrue(self.gui.get_active())
        self.message_broker.publish(PresetData(None, None, autoload=False))
        self.assertFalse(self.gui.get_active())

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(PresetData(None, None, autoload=True))
        self.message_broker.publish(PresetData(None, None, autoload=False))
        self.controller_mock.set_autoload.assert_not_called()


@test_setup
class TestReleaseCombinationSwitch(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Switch()
        self.switch = ReleaseCombinationSwitch(
            self.message_broker, self.controller_mock, self.gui
        )

    def test_updates_mapping(self):
        self.gui.set_active(True)
        self.controller_mock.update_mapping.assert_called_once_with(
            release_combination_keys=True
        )
        self.controller_mock.reset_mock()
        self.gui.set_active(False)
        self.controller_mock.update_mapping.assert_called_once_with(
            release_combination_keys=False
        )

    def test_updates_state(self):
        self.message_broker.publish(MappingData(release_combination_keys=True))
        self.assertTrue(self.gui.get_active())
        self.message_broker.publish(MappingData(release_combination_keys=False))
        self.assertFalse(self.gui.get_active())

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(MappingData(release_combination_keys=True))
        self.message_broker.publish(MappingData(release_combination_keys=False))
        self.controller_mock.update_mapping.assert_not_called()


@test_setup
class TestEventEntry(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = InputConfigEntry(
            InputConfig(type=3, code=0, analog_threshold=1), self.controller_mock
        )

    def test_move_event(self):
        self.gui._up_btn.clicked()
        self.controller_mock.move_input_config_in_combination.assert_called_once_with(
            InputConfig(type=3, code=0, analog_threshold=1), "up"
        )
        self.controller_mock.reset_mock()

        self.gui._down_btn.clicked()
        self.controller_mock.move_input_config_in_combination.assert_called_once_with(
            InputConfig(type=3, code=0, analog_threshold=1), "down"
        )


@test_setup
class TestCombinationListbox(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.ListBox()
        self.listbox = CombinationListbox(
            self.message_broker, self.controller_mock, self.gui
        )
        self.controller_mock.is_empty_mapping.return_value = False
        combination = InputCombination(
            (
                InputConfig(type=1, code=1),
                InputConfig(type=3, code=0, analog_threshold=1),
                InputConfig(type=1, code=2),
            )
        )
        self.message_broker.publish(
            MappingData(
                input_combination=combination.to_config(), target_uinput="keyboard"
            )
        )

    def get_selected_row(self) -> InputConfigEntry:
        for entry in self.gui.get_children():
            if entry.is_selected():
                return entry

        raise Exception("Expected one InputConfigEntry to be selected")

    def select_row(self, input_cfg: InputConfig):
        for entry in self.gui.get_children():
            if entry.input_event == input_cfg:
                self.gui.select_row(entry)

    def test_loads_selected_row(self):
        self.select_row(InputConfig(type=1, code=2))
        self.controller_mock.load_input_config.assert_called_once_with(
            InputConfig(type=1, code=2)
        )

    def test_does_not_create_rows_when_mapping_is_empty(self):
        self.controller_mock.is_empty_mapping.return_value = True
        combination = InputCombination(
            (
                InputConfig(type=1, code=1),
                InputConfig(type=3, code=0, analog_threshold=1),
            )
        )
        self.message_broker.publish(MappingData(input_combination=combination))
        self.assertEqual(len(self.gui.get_children()), 0)

    def test_selects_row_when_selected_event_message_arrives(self):
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=1))
        self.assertEqual(
            self.get_selected_row().input_event,
            InputConfig(type=3, code=0, analog_threshold=1),
        )

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=1))
        self.controller_mock.load_event.assert_not_called()


@test_setup
class TestAnalogInputSwitch(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Switch()
        self.switch = AnalogInputSwitch(
            self.message_broker, self.controller_mock, self.gui
        )

    def test_updates_event_as_analog(self):
        self.gui.set_active(True)
        self.controller_mock.set_event_as_analog.assert_called_once_with(True)
        self.controller_mock.reset_mock()
        self.gui.set_active(False)
        self.controller_mock.set_event_as_analog.assert_called_once_with(False)

    def test_updates_state(self):
        self.message_broker.publish(InputConfig(type=3, code=0))
        self.assertTrue(self.gui.get_active())
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=10))
        self.assertFalse(self.gui.get_active())

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(InputConfig(type=3, code=0))
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=-10))
        self.controller_mock.set_event_as_analog.assert_not_called()

    def test_disables_switch_when_key_event(self):
        self.message_broker.publish(InputConfig(type=1, code=1))
        self.assertLess(self.gui.get_opacity(), 0.6)
        self.assertFalse(self.gui.get_sensitive())

    def test_enables_switch_when_axis_event(self):
        self.message_broker.publish(InputConfig(type=1, code=1))
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=10))
        self.assertEqual(self.gui.get_opacity(), 1)
        self.assertTrue(self.gui.get_sensitive())

        self.message_broker.publish(InputConfig(type=1, code=1))
        self.message_broker.publish(InputConfig(type=2, code=0, analog_threshold=10))
        self.assertEqual(self.gui.get_opacity(), 1)
        self.assertTrue(self.gui.get_sensitive())


@test_setup
class TestTriggerThresholdInput(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.SpinButton()
        self.input = TriggerThresholdInput(
            self.message_broker, self.controller_mock, self.gui
        )
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=-10))

    def assert_abs_event_config(self):
        self.assertEqual(self.gui.get_range(), (-99, 99))
        self.assertTrue(self.gui.get_sensitive())
        self.assertEqual(self.gui.get_opacity(), 1)

    def assert_rel_event_config(self):
        self.assertEqual(self.gui.get_range(), (-999, 999))
        self.assertTrue(self.gui.get_sensitive())
        self.assertEqual(self.gui.get_opacity(), 1)

    def assert_key_event_config(self):
        self.assertFalse(self.gui.get_sensitive())
        self.assertLess(self.gui.get_opacity(), 0.6)

    def test_updates_event(self):
        self.gui.set_value(15)
        self.controller_mock.update_input_config.assert_called_once_with(
            InputConfig(type=3, code=0, analog_threshold=15)
        )

    def test_sets_value_on_selected_event_message(self):
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=10))
        self.assertEqual(self.gui.get_value(), 10)

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(InputConfig(type=3, code=0, analog_threshold=10))
        self.controller_mock.update_input_config.assert_not_called()

    def test_updates_configuration_according_to_selected_event(self):
        self.assert_abs_event_config()
        self.message_broker.publish(InputConfig(type=2, code=0, analog_threshold=-10))
        self.assert_rel_event_config()
        self.message_broker.publish(InputConfig(type=1, code=1))
        self.assert_key_event_config()


@test_setup
class TestReleaseTimeoutInput(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.SpinButton()
        self.input = ReleaseTimeoutInput(
            self.message_broker, self.controller_mock, self.gui
        )
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    [InputConfig(type=2, code=0, analog_threshold=1)]
                ),
                target_uinput="keyboard",
            )
        )

    def test_updates_timeout_on_mapping_message(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    [InputConfig(type=2, code=0, analog_threshold=1)]
                ),
                release_timeout=1,
            )
        )
        self.assertEqual(self.gui.get_value(), 1)

    def test_updates_mapping(self):
        self.gui.set_value(0.5)
        self.controller_mock.update_mapping.assert_called_once_with(release_timeout=0.5)

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    [InputConfig(type=2, code=0, analog_threshold=1)]
                ),
                release_timeout=1,
            )
        )
        self.controller_mock.update_mapping.assert_not_called()

    def test_disables_input_based_on_input_combination(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=2, code=0, analog_threshold=1),
                        InputConfig(type=1, code=1),
                    )
                )
            )
        )
        self.assertTrue(self.gui.get_sensitive())
        self.assertEqual(self.gui.get_opacity(), 1)

        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=1, code=1),
                        InputConfig(type=1, code=2),
                    )
                )
            )
        )
        self.assertFalse(self.gui.get_sensitive())
        self.assertLess(self.gui.get_opacity(), 0.6)

        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=2, code=0, analog_threshold=1),
                        InputConfig(type=1, code=1),
                    )
                )
            )
        )
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    (
                        InputConfig(type=3, code=0, analog_threshold=1),
                        InputConfig(
                            type=1,
                            code=2,
                        ),
                    )
                )
            )
        )
        self.assertFalse(self.gui.get_sensitive())
        self.assertLess(self.gui.get_opacity(), 0.6)


@test_setup
class TestOutputAxisSelector(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.ComboBox()
        self.selection = OutputAxisSelector(
            self.message_broker, self.controller_mock, self.gui
        )
        absinfo = evdev.AbsInfo(0, -10, 10, 0, 0, 0)
        self.message_broker.publish(
            UInputsData(
                {
                    "mouse": {1: [1, 2, 3, 4], 2: [0, 1, 2, 3]},
                    "keyboard": {1: [1, 2, 3, 4]},
                    "gamepad": {
                        2: [0, 1, 2, 3],
                        3: [(0, absinfo), (1, absinfo), (2, absinfo), (3, absinfo)],
                    },
                }
            )
        )
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=1, code=1)]),
            )
        )

    def set_active_selection(self, selection: Tuple):
        self.gui.set_active_id(f"{selection[0]}, {selection[1]}")

    def get_active_selection(self) -> Tuple[int, int]:
        return tuple(int(i) for i in self.gui.get_active_id().split(","))  # type: ignore

    def test_updates_mapping(self):
        self.set_active_selection((2, 0))
        self.controller_mock.update_mapping.assert_called_once_with(
            output_type=2, output_code=0
        )

    def test_updates_mapping_with_none(self):
        self.set_active_selection((2, 0))
        self.controller_mock.reset_mock()
        self.set_active_selection((None, None))
        self.controller_mock.update_mapping.assert_called_once_with(
            output_type=None, output_code=None
        )

    def test_selects_correct_entry(self):
        self.assertEqual(self.gui.get_active_id(), "None, None")
        self.message_broker.publish(
            MappingData(target_uinput="mouse", output_type=2, output_code=3)
        )
        self.assertEqual(self.get_active_selection(), (2, 3))

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(
            MappingData(target_uinput="mouse", output_type=2, output_code=3)
        )
        self.controller_mock.update_mapping.assert_not_called()

    def test_updates_dropdown_model(self):
        self.assertEqual(len(self.gui.get_model()), 5)
        self.message_broker.publish(MappingData(target_uinput="keyboard"))
        self.assertEqual(len(self.gui.get_model()), 1)
        self.message_broker.publish(MappingData(target_uinput="gamepad"))
        self.assertEqual(len(self.gui.get_model()), 9)


@test_setup
class TestKeyAxisStackSwitcher(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Box()
        self.gtk_stack = Gtk.Stack()
        self.analog_toggle = Gtk.ToggleButton()
        self.key_toggle = Gtk.ToggleButton()

        self.gui.add(self.gtk_stack)
        self.gui.add(self.analog_toggle)
        self.gui.add(self.key_toggle)
        self.gtk_stack.add_named(Gtk.Box(), "Analog Axis")
        self.gtk_stack.add_named(Gtk.Box(), "Key or Macro")

        self.stack = KeyAxisStackSwitcher(
            self.message_broker,
            self.controller_mock,
            self.gtk_stack,
            self.key_toggle,
            self.analog_toggle,
        )

        self.gui.show_all()
        self.gtk_stack.set_visible_child_name("Key or Macro")

    def assert_key_macro_active(self):
        self.assertEqual(self.gtk_stack.get_visible_child_name(), "Key or Macro")
        self.assertTrue(self.key_toggle.get_active())
        self.assertFalse(self.analog_toggle.get_active())

    def assert_analog_active(self):
        self.assertEqual(self.gtk_stack.get_visible_child_name(), "Analog Axis")
        self.assertFalse(self.key_toggle.get_active())
        self.assertTrue(self.analog_toggle.get_active())

    def test_switches_to_axis(self):
        self.message_broker.publish(MappingData(mapping_type="analog"))
        self.assert_analog_active()

    def test_switches_to_key_macro(self):
        self.message_broker.publish(MappingData(mapping_type="analog"))
        self.message_broker.publish(MappingData(mapping_type="key_macro"))
        self.assert_key_macro_active()

    def test_updates_mapping_type(self):
        self.key_toggle.set_active(True)
        self.controller_mock.update_mapping.assert_called_once_with(
            mapping_type="key_macro"
        )
        self.controller_mock.update_mapping.reset_mock()

        self.analog_toggle.set_active(True)
        self.controller_mock.update_mapping.assert_called_once_with(
            mapping_type="analog"
        )

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(MappingData(mapping_type="analog"))
        self.message_broker.publish(MappingData(mapping_type="key_macro"))
        self.controller_mock.update_mapping.assert_not_called()


@test_setup
class TestTransformationDrawArea(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Window()
        self.draw_area = Gtk.DrawingArea()
        self.gui.add(self.draw_area)
        self.transform_draw_area = TransformationDrawArea(
            self.message_broker,
            self.controller_mock,
            self.draw_area,
        )

    def test_draws_transform(self):
        with spy(self.transform_draw_area, "_transformation") as mock:
            # show the window, it takes some time and iterations until it pops up
            self.gui.show_all()
            for _ in range(5):
                gtk_iteration()
                time.sleep(0.01)

            mock.assert_called()

    def test_updates_transform_when_mapping_updates(self):
        old_tf = self.transform_draw_area._transformation
        self.message_broker.publish(MappingData(gain=2))
        self.assertIsNot(old_tf, self.transform_draw_area._transformation)

    def test_redraws_when_mapping_updates(self):
        self.gui.show_all()
        gtk_iteration(20)
        mock = MagicMock()
        self.draw_area.connect("draw", mock)
        self.message_broker.publish(MappingData(gain=2))
        gtk_iteration(20)
        mock.assert_called()


@test_setup
class TestSliders(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.Box()
        self.gain = Gtk.Scale()
        self.deadzone = Gtk.Scale()
        self.expo = Gtk.Scale()

        # add everything to a box: it will be cleand up properly
        self.gui.add(self.gain)
        self.gui.add(self.deadzone)
        self.gui.add(self.expo)

        self.sliders = Sliders(
            self.message_broker,
            self.controller_mock,
            self.gain,
            self.deadzone,
            self.expo,
        )
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination([InputConfig(type=3, code=0)]),
                target_uinput="mouse",
            )
        )

    @staticmethod
    def get_range(range: Gtk.Range) -> Tuple[int, int]:
        """the Gtk.Range, has no get_range method. this is a workaround"""
        v = range.get_value()
        range.set_value(-(2**16))
        min_ = range.get_value()
        range.set_value(2**16)
        max_ = range.get_value()
        range.set_value(v)
        return min_, max_

    def test_slider_ranges(self):
        self.assertEqual(self.get_range(self.gain), (-2, 2))
        self.assertEqual(self.get_range(self.deadzone), (0, 0.9))
        self.assertEqual(self.get_range(self.expo), (-1, 1))

    def test_updates_value(self):
        self.message_broker.publish(
            MappingData(
                gain=0.5,
                deadzone=0.6,
                expo=0.3,
            )
        )
        self.assertEqual(self.gain.get_value(), 0.5)
        self.assertEqual(self.expo.get_value(), 0.3)
        self.assertEqual(self.deadzone.get_value(), 0.6)

    def test_gain_updates_mapping(self):
        self.gain.set_value(0.5)
        self.controller_mock.update_mapping.assert_called_once_with(gain=0.5)

    def test_expo_updates_mapping(self):
        self.expo.set_value(0.5)
        self.controller_mock.update_mapping.assert_called_once_with(expo=0.5)

    def test_deadzone_updates_mapping(self):
        self.deadzone.set_value(0.5)
        self.controller_mock.update_mapping.assert_called_once_with(deadzone=0.5)

    def test_avoids_recursion(self):
        self.message_broker.publish(MappingData(gain=0.5))
        self.controller_mock.update_mapping.assert_not_called()
        self.message_broker.publish(MappingData(expo=0.5))
        self.controller_mock.update_mapping.assert_not_called()
        self.message_broker.publish(MappingData(deadzone=0.5))
        self.controller_mock.update_mapping.assert_not_called()


@test_setup
class TestRelativeInputCutoffInput(ComponentBaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.gui = Gtk.SpinButton()
        self.input = RelativeInputCutoffInput(
            self.message_broker, self.controller_mock, self.gui
        )
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=2, code=0)]),
                rel_to_abs_input_cutoff=1,
                output_type=3,
                output_code=0,
            )
        )

    def assert_active(self):
        self.assertTrue(self.gui.get_sensitive())
        self.assertEqual(self.gui.get_opacity(), 1)

    def assert_inactive(self):
        self.assertFalse(self.gui.get_sensitive())
        self.assertLess(self.gui.get_opacity(), 0.6)

    def test_avoids_infinite_recursion(self):
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=2, code=0)]),
                rel_to_abs_input_cutoff=3,
                output_type=3,
                output_code=0,
            )
        )
        self.controller_mock.update_mapping.assert_not_called()

    def test_updates_value(self):
        rel_to_abs_input_cutoff = 3
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=2, code=0)]),
                rel_to_abs_input_cutoff=rel_to_abs_input_cutoff,
                output_type=3,
                output_code=0,
            )
        )
        self.assertEqual(self.gui.get_value(), rel_to_abs_input_cutoff)

    def test_updates_mapping(self):
        self.gui.set_value(300)
        self.controller_mock.update_mapping.assert_called_once_with(rel_xy_cutoff=300)

    def test_disables_input_when_no_rel_axis_input(self):
        self.assert_active()
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=3, code=0)]),
                output_type=3,
                output_code=0,
            )
        )
        self.assert_inactive()

    def test_disables_input_when_no_abs_axis_output(self):
        self.assert_active()
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=2, code=0)]),
                rel_to_abs_input_cutoff=3,
                output_type=2,
                output_code=0,
            )
        )
        self.assert_inactive()

    def test_enables_input(self):
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=3, code=0)]),
                output_type=3,
                output_code=0,
            )
        )
        self.assert_inactive()
        self.message_broker.publish(
            MappingData(
                target_uinput="mouse",
                input_combination=InputCombination([InputConfig(type=2, code=0)]),
                rel_to_abs_input_cutoff=1,
                output_type=3,
                output_code=0,
            )
        )
        self.assert_active()


@test_setup
class TestRequireActiveMapping(ComponentBaseTest):
    def test_no_reqorded_input_required(self):
        self.box = Gtk.Box()
        RequireActiveMapping(
            self.message_broker,
            self.box,
            require_recorded_input=False,
        )
        combination = InputCombination([InputConfig(type=1, code=KEY_A)])

        self.message_broker.publish(MappingData())
        self.assert_inactive(self.box)

        self.message_broker.publish(PresetData(name="preset", mappings=()))
        self.assert_inactive(self.box)

        # a mapping is available, that is all the widget needs to be activated. one
        # mapping is always selected, so there is no need to check the mapping message
        self.message_broker.publish(PresetData(name="preset", mappings=(combination,)))
        self.assert_active(self.box)

        self.message_broker.publish(MappingData(input_combination=combination))
        self.assert_active(self.box)

        self.message_broker.publish(MappingData())
        self.assert_active(self.box)

    def test_recorded_input_required(self):
        self.box = Gtk.Box()
        RequireActiveMapping(
            self.message_broker,
            self.box,
            require_recorded_input=True,
        )
        combination = InputCombination([InputConfig(type=1, code=KEY_A)])

        self.message_broker.publish(MappingData())
        self.assert_inactive(self.box)

        self.message_broker.publish(PresetData(name="preset", mappings=()))
        self.assert_inactive(self.box)

        self.message_broker.publish(PresetData(name="preset", mappings=(combination,)))
        self.assert_inactive(self.box)

        # the widget will be enabled once a mapping with recorded input is selected
        self.message_broker.publish(MappingData(input_combination=combination))
        self.assert_active(self.box)

        # this mapping doesn't have input recorded, so the box is disabled
        self.message_broker.publish(MappingData())
        self.assert_inactive(self.box)

    def assert_inactive(self, widget: Gtk.Widget):
        self.assertFalse(widget.get_sensitive())
        self.assertLess(widget.get_opacity(), 0.6)
        self.assertGreater(widget.get_opacity(), 0.4)

    def assert_active(self, widget: Gtk.Widget):
        self.assertTrue(widget.get_sensitive())
        self.assertEqual(widget.get_opacity(), 1)


@test_setup
class TestStack(ComponentBaseTest):
    def test_switches_pages(self):
        self.stack = Gtk.Stack()
        self.stack.add_named(Gtk.Label(), "Devices")
        self.stack.add_named(Gtk.Label(), "Presets")
        self.stack.add_named(Gtk.Label(), "Editor")
        self.stack.show_all()
        stack_wrapper = Stack(self.message_broker, self.controller_mock, self.stack)

        self.message_broker.publish(DoStackSwitch(Stack.devices_page))
        self.assertEqual(self.stack.get_visible_child_name(), "Devices")

        self.message_broker.publish(DoStackSwitch(Stack.presets_page))
        self.assertEqual(self.stack.get_visible_child_name(), "Presets")

        self.message_broker.publish(DoStackSwitch(Stack.editor_page))
        self.assertEqual(self.stack.get_visible_child_name(), "Editor")


@test_setup
class TestBreadcrumbs(ComponentBaseTest):
    def test_breadcrumbs(self):
        self.label_1 = Gtk.Label()
        self.label_2 = Gtk.Label()
        self.label_3 = Gtk.Label()
        self.label_4 = Gtk.Label()
        self.label_5 = Gtk.Label()

        Breadcrumbs(
            self.message_broker,
            self.label_1,
            show_device_group=False,
            show_preset=False,
            show_mapping=False,
        )
        Breadcrumbs(
            self.message_broker,
            self.label_2,
            show_device_group=True,
            show_preset=False,
            show_mapping=False,
        )
        Breadcrumbs(
            self.message_broker,
            self.label_3,
            show_device_group=True,
            show_preset=True,
            show_mapping=False,
        )
        Breadcrumbs(
            self.message_broker,
            self.label_4,
            show_device_group=True,
            show_preset=True,
            show_mapping=True,
        )
        Breadcrumbs(
            self.message_broker,
            self.label_5,
            show_device_group=False,
            show_preset=False,
            show_mapping=True,
        )

        self.assertEqual(self.label_1.get_text(), "")
        self.assertEqual(self.label_2.get_text(), "?")
        self.assertEqual(self.label_3.get_text(), "?  /  ?")
        self.assertEqual(self.label_4.get_text(), "?  /  ?  /  ?")
        self.assertEqual(self.label_5.get_text(), "?")

        self.message_broker.publish(PresetData("preset", None))

        self.assertEqual(self.label_1.get_text(), "")
        self.assertEqual(self.label_2.get_text(), "?")
        self.assertEqual(self.label_3.get_text(), "?  /  preset")
        self.assertEqual(self.label_4.get_text(), "?  /  preset  /  ?")
        self.assertEqual(self.label_5.get_text(), "?")

        self.message_broker.publish(GroupData("group", ()))

        self.assertEqual(self.label_1.get_text(), "")
        self.assertEqual(self.label_2.get_text(), "group")
        self.assertEqual(self.label_3.get_text(), "group  /  preset")
        self.assertEqual(self.label_4.get_text(), "group  /  preset  /  ?")
        self.assertEqual(self.label_5.get_text(), "?")

        self.message_broker.publish(MappingData())

        self.assertEqual(self.label_1.get_text(), "")
        self.assertEqual(self.label_2.get_text(), "group")
        self.assertEqual(self.label_3.get_text(), "group  /  preset")
        self.assertEqual(self.label_4.get_text(), "group  /  preset  /  Empty Mapping")
        self.assertEqual(self.label_5.get_text(), "Empty Mapping")

        self.message_broker.publish(MappingData(name="mapping"))
        self.assertEqual(self.label_4.get_text(), "group  /  preset  /  mapping")
        self.assertEqual(self.label_5.get_text(), "mapping")

        combination = InputCombination(
            (
                InputConfig(type=1, code=KEY_A),
                InputConfig(type=1, code=KEY_B),
            )
        )
        self.message_broker.publish(MappingData(input_combination=combination))
        self.assertEqual(self.label_4.get_text(), "group  /  preset  /  a + b")
        self.assertEqual(self.label_5.get_text(), "a + b")

        combination = InputCombination([InputConfig(type=1, code=KEY_A)])
        self.message_broker.publish(
            MappingData(name="qux", input_combination=combination)
        )
        self.assertEqual(self.label_4.get_text(), "group  /  preset  /  qux")
        self.assertEqual(self.label_5.get_text(), "qux")



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Optional

from gi.repository import Gtk

from inputremapper.gui.components.common import FlowBoxEntry, FlowBoxWrapper
from inputremapper.gui.components.editor import ICON_PRIORITIES, ICON_NAMES
from inputremapper.gui.components.main import Stack
from inputremapper.gui.controller import Controller
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import (
    GroupsData,
    GroupData,
    DoStackSwitch,
)
from inputremapper.logging.logger import logger


class DeviceGroupEntry(FlowBoxEntry):
    """A device that can be selected in the GUI.

    For example a keyboard or a mouse.
    """

    __gtype_name__ = "DeviceGroupEntry"

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        icon_name: Optional[str],
        group_key: str,
    ):
        super().__init__(
            message_broker=message_broker,
            controller=controller,
            icon_name=icon_name,
            name=group_key,
        )
        self.group_key = group_key

    def _on_gtk_toggle(self, *_, **__):
        logger.debug('Selecting device "%s"', self.group_key)
        self._controller.load_group(self.group_key)
        self.message_broker.publish(DoStackSwitch(Stack.presets_page))


class DeviceGroupSelection(FlowBoxWrapper):
    """A wrapper for the container with our groups.

    A group is a collection of devices.
    """

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        flowbox: Gtk.FlowBox,
    ):
        super().__init__(flowbox)

        self._message_broker = message_broker
        self._controller = controller
        self._gui = flowbox

        self._message_broker.subscribe(MessageType.groups, self._on_groups_changed)
        self._message_broker.subscribe(MessageType.group, self._on_group_changed)

    def _on_groups_changed(self, data: GroupsData):
        self._gui.foreach(self._gui.remove)

        for group_key, types in data.groups.items():
            if len(types) > 0:
                device_type = sorted(types, key=ICON_PRIORITIES.index)[0]
                icon_name = ICON_NAMES[device_type]
            else:
                icon_name = None

            logger.debug("adding %s to device selection", group_key)
            device_group_entry = DeviceGroupEntry(
                self._message_broker,
                self._controller,
                icon_name,
                group_key,
            )
            self._gui.insert(device_group_entry, -1)

        if self._controller.data_manager.active_group:
            self.show_active_entry(self._controller.data_manager.active_group.key)

    def _on_group_changed(self, data: GroupData):
        self.show_active_entry(data.group_key)



import asyncio
import time
import unittest
from unittest.mock import patch

import evdev
from evdev.ecodes import KEY_A, EV_KEY, KEY_B, KEY_LEFTSHIFT, KEY_C

from inputremapper.configs.input_config import InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.configs.preset import Preset
from inputremapper.injection.context import Context
from inputremapper.injection.event_reader import EventReader
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.macros.parse import Parser
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.input_event import InputEvent
from tests.lib.fixtures import fixtures
from tests.lib.patches import InputDevice
from tests.lib.pipes import uinput_write_history
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestModTapIntegration(unittest.IsolatedAsyncioTestCase):
    # Testcases are from https://github.com/qmk/qmk_firmware/blob/78a0adfbb4d2c4e12f93f2a62ded0020d406243e/docs/tap_hold.md#nested-tap-abba-nested-tap
    # This test-setup is a bit more involved, because I needed to modify the
    # EventReader as well for this to work.

    def setUp(self):
        self.origin_hash = fixtures.bar_device.get_device_hash()
        self.forward_uinput = evdev.UInput(name="test-forward-uinput")
        self.source_device = InputDevice(fixtures.bar_device.path)
        self.stop_event = asyncio.Event()
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.target_uinput = self.global_uinputs.get_uinput("keyboard")
        self.mapping_parser = MappingParser(self.global_uinputs)

        self.mapping = Mapping.from_combination(
            input_combination=[
                InputConfig(
                    type=EV_KEY,
                    code=KEY_A,
                    origin_hash=self.origin_hash,
                )
            ],
            output_symbol="mod_tap(a, Shift_L)",
        )

        self.preset = Preset()
        self.preset.add(self.mapping)

        self.bootstrap_event_reader()

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def bootstrap_event_reader(self):
        self.context = Context(
            self.preset,
            source_devices={self.origin_hash: self.source_device},
            forward_devices={self.origin_hash: self.forward_uinput},
            mapping_parser=self.mapping_parser,
        )

        self.event_reader = EventReader(
            self.context,
            self.source_device,
            self.stop_event,
        )

    def write(self, type_, code, value):
        self.target_uinput.write_event(InputEvent.from_tuple((type_, code, value)))

    async def input(self, type_, code, value):
        asyncio.ensure_future(
            self.event_reader.handle(
                InputEvent.from_tuple(
                    (
                        type_,
                        code,
                        value,
                    ),
                    origin_hash=self.origin_hash,
                )
            )
        )
        # Make the main_loop iterate a bit for the event_reader to do its thing.
        await asyncio.sleep(0)

    async def test_distinct_taps_1(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.190)
        await self.input(EV_KEY, KEY_A, 0)
        await asyncio.sleep(0.020)  # exceeds tapping_term here
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)

        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_A, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_A, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )
        self.assertEqual(
            self.target_uinput.write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_A, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_A, 0)),
            ],
        )
        self.assertEqual(
            self.forward_uinput.write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )

    async def test_distinct_taps_2(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.220)  # exceeds tapping_term here
        await self.input(EV_KEY, KEY_A, 0)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)

        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )
        self.assertEqual(
            self.target_uinput.write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 0)),
            ],
        )
        self.assertEqual(
            self.forward_uinput.write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )

    async def test_nested_tap_1(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.100)
        self.assertEqual(uinput_write_history, [])

        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        self.assertEqual(uinput_write_history, [])

        await self.input(EV_KEY, KEY_B, 0)
        await asyncio.sleep(0.050)
        self.assertEqual(uinput_write_history, [])

        await self.input(EV_KEY, KEY_A, 0)

        # everything happened within the tapping_term, so the modifier is not activated.
        # "ab" should be written, in the exact order of the input.
        await asyncio.sleep(0.040)
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_A, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_A, 0)),
            ],
        )

    async def test_nested_tap_2(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.100)
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)
        await asyncio.sleep(0.100)  # exceeds tapping_term here
        await self.input(EV_KEY, KEY_A, 0)

        await asyncio.sleep(0.020)
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 0)),
            ],
        )

    async def test_nested_tap_3(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.220)  # exceeds tapping_term here
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_A, 0)

        await asyncio.sleep(0.020)
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 0)),
            ],
        )

    async def test_rolling_keys_1(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.100)
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_A, 0)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)

        # everything happened within the tapping_term, so the modifier is not activated.
        # "ab" should be written, in the exact order of the input.
        await asyncio.sleep(0.100)
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_A, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_A, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )

    async def test_rolling_keys_2(self):
        await self.input(EV_KEY, KEY_A, 1)
        await asyncio.sleep(0.100)
        await self.input(EV_KEY, KEY_B, 1)
        await asyncio.sleep(0.100)  # exceeds tapping_term here
        await self.input(EV_KEY, KEY_A, 0)
        await asyncio.sleep(0.020)
        await self.input(EV_KEY, KEY_B, 0)

        await asyncio.sleep(0.020)
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 1)),
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 0)),
                InputEvent.from_tuple((EV_KEY, KEY_B, 0)),
            ],
        )

    async def test_many_keys_correct_order_without_sleep(self):
        self.mapping.macro_key_sleep_ms = 0
        await self.many_keys_correct_order()

    async def test_many_keys_correct_order_with_sleep(self):
        self.mapping.macro_key_sleep_ms = 20
        await self.many_keys_correct_order()

    async def many_keys_correct_order(self):
        await self.input(EV_KEY, KEY_A, 1)

        # Send many events to the listener. It has to make all of them wait.
        for i in range(30):
            await self.input(EV_KEY, i, 1)

        # exceed tapping_term. mod_tap will inject the modifier and replay all the
        # previous events.
        await asyncio.sleep(0.201)

        # mod_tap is busy replaying events. While it does that, inject this
        await self.input(EV_KEY, 100, 1)

        start = time.time()
        timeout = 2
        while len(uinput_write_history) < 32 and (time.time() - start) < timeout:
            # Wait for it to complete
            await asyncio.sleep(0.1)

        self.assertEqual(len(uinput_write_history), 32)

        # Expect it to cleanly handle all events before injecting 100. Expect
        # everything to be in the correct order.
        self.assertEqual(
            uinput_write_history,
            [
                InputEvent.from_tuple((EV_KEY, KEY_LEFTSHIFT, 1)),
                *[InputEvent.from_tuple((EV_KEY, i, 1)) for i in range(30)],
                InputEvent.from_tuple((EV_KEY, 100, 1)),
            ],
        )

    async def test_mapped_second_key(self):
        # Map b to c.
        # While mod_tap is waiting for the timeout to happen, press b.
        # We expect c to be written, because b goes through the handlers and
        # gets mapped.
        # The event_reader has to wait for listeners to complete for mod_tap to work, so
        # that it hands them over to the other handlers when the time comes.
        # That means however, that the event_readers loop blocks. Therefore, it was turned
        # into a fire-and-forget kind of thing. When an event arrives, it just schedules
        # asyncio to do that stuff later, and continues reading.

        self.preset.add(
            Mapping.from_combination(
                input_combination=[
                    InputConfig(
                        type=EV_KEY,
                        code=KEY_B,
                        origin_hash=self.origin_hash,
                    )
                ],
                output_symbol="c",
            ),
        )

        self.bootstrap_event_reader()

        async def async_generator():
            events = [
                InputEvent(0, 0, EV_KEY, KEY_A, 1),
                InputEvent(0, 0, EV_KEY, KEY_B, 1),
                InputEvent(0, 0, EV_KEY, KEY_A, 0),
                InputEvent(0, 0, EV_KEY, KEY_B, 0),
            ]
            for event in events:
                yield event
                # Wait a bit. During runtime, events don't come in that quickly
                # and the mod_tap macro needs some loop iterations until it adds
                # the listener to the context.
                await asyncio.sleep(0.010)

        with patch.object(self.event_reader, "read_loop", async_generator):
            await self.event_reader.run()

        await asyncio.sleep(0.020)
        self.assertIn(InputEvent(0, 0, EV_KEY, KEY_C, 1), uinput_write_history)
        self.assertIn(InputEvent(0, 0, EV_KEY, KEY_C, 0), uinput_write_history)
        self.assertIn(InputEvent(0, 0, EV_KEY, KEY_A, 1), uinput_write_history)
        self.assertIn(InputEvent(0, 0, EV_KEY, KEY_A, 0), uinput_write_history)
        self.assertNotIn(InputEvent(0, 0, EV_KEY, KEY_B, 1), uinput_write_history)
        self.assertNotIn(InputEvent(0, 0, EV_KEY, KEY_B, 0), uinput_write_history)


@test_setup
class TestModTapUnit(MacroTestBase):
    async def wait_for_timeout(self, macro):
        macro = Parser.parse(macro, self.context, DummyMapping, True)

        start = time.time()
        # Awaiting macro.run will cause it to wait for the tapping_term.
        # When it injects the modifier, release the trigger.
        macro.press_trigger()
        await macro.run(lambda *_, **__: macro.release_trigger())
        return time.time() - start

    async def test_tapping_term_configuration_default(self):
        time_ = await self.wait_for_timeout("mod_tap(a, b)")
        # + 2 times 10ms of keycode_pause
        self.assertAlmostEqual(time_, 0.22, delta=0.02)

    async def test_tapping_term_configuration_100(self):
        time_ = await self.wait_for_timeout("mod_tap(a, b, 100)")
        self.assertAlmostEqual(time_, 0.12, delta=0.02)

    async def test_tapping_term_configuration_100_kwarg(self):
        time_ = await self.wait_for_timeout("mod_tap(a, b, tapping_term=100)")
        self.assertAlmostEqual(time_, 0.12, delta=0.02)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Keeps injecting keycodes in the background based on the preset."""
from __future__ import annotations

import asyncio
import enum
import multiprocessing
import sys
import time
from collections import defaultdict
from dataclasses import dataclass
from multiprocessing.connection import Connection
from typing import Dict, List, Optional, Tuple, Union

import evdev

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.preset import Preset
from inputremapper.groups import (
    _Group,
    classify,
    DeviceType,
)
from inputremapper.gui.messages.message_broker import MessageType
from inputremapper.injection.context import Context
from inputremapper.injection.event_reader import EventReader
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.injection.numlock import set_numlock, is_numlock_on, ensure_numlock
from inputremapper.logging.logger import logger
from inputremapper.utils import get_device_hash, DeviceHash

CapabilitiesDict = Dict[int, List[int]]

DEV_NAME = "input-remapper"


# messages sent to the injector process
class InjectorCommand(str, enum.Enum):
    CLOSE = "CLOSE"


# messages the injector process reports back to the service
class InjectorState(str, enum.Enum):
    UNKNOWN = "UNKNOWN"
    STARTING = "STARTING"
    ERROR = "FAILED"
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    NO_GRAB = "NO_GRAB"
    UPGRADE_EVDEV = "UPGRADE_EVDEV"


def is_in_capabilities(
    combination: InputCombination, capabilities: CapabilitiesDict
) -> bool:
    """Are this combination or one of its sub keys in the capabilities?"""
    for event in combination:
        if event.code in capabilities.get(event.type, []):
            return True

    return False


def get_udev_name(name: str, suffix: str) -> str:
    """Make sure the generated name is not longer than 80 chars."""
    max_len = 80  # based on error messages
    remaining_len = max_len - len(DEV_NAME) - len(suffix) - 2
    middle = name[:remaining_len]
    name = f"{DEV_NAME} {middle} {suffix}"
    return name


@dataclass(frozen=True)
class InjectorStateMessage:
    message_type = MessageType.injector_state
    state: Union[InjectorState]

    def active(self) -> bool:
        return self.state in [InjectorState.RUNNING, InjectorState.STARTING]

    def inactive(self) -> bool:
        return self.state in [InjectorState.STOPPED, InjectorState.NO_GRAB]


class Injector(multiprocessing.Process):
    """Initializes, starts and stops injections.

    Is a process to make it non-blocking for the rest of the code and to
    make running multiple injector easier. There is one process per
    hardware-device that is being mapped.
    """

    group: _Group
    preset: Preset
    context: Optional[Context]
    _devices: List[evdev.InputDevice]
    _state: InjectorState
    _msg_pipe: Tuple[Connection, Connection]
    _event_readers: List[EventReader]
    _stop_event: asyncio.Event

    regrab_timeout = 0.2

    def __init__(
        self,
        group: _Group,
        preset: Preset,
        mapping_parser: MappingParser,
    ) -> None:
        """

        Parameters
        ----------
        group
            the device group
        """
        self.group = group
        self.mapping_parser = mapping_parser
        self._state = InjectorState.UNKNOWN

        # used to interact with the parts of this class that are running within
        # the new process
        self._msg_pipe = multiprocessing.Pipe()

        self.preset = preset
        self.context = None  # only needed inside the injection process

        self._event_readers = []

        super().__init__(name=group.key)

    """Functions to interact with the running process."""

    def get_state(self) -> InjectorState:
        """Get the state of the injection.

        Can be safely called from the main process.
        """
        # before we try to we try to guess anything lets check if there is a message
        state = self._state
        while self._msg_pipe[1].poll():
            state = self._msg_pipe[1].recv()

        # figure out what is going on step by step
        alive = self.is_alive()

        # if `self.start()` has been called
        started = state != InjectorState.UNKNOWN or alive

        if started:
            if state == InjectorState.UNKNOWN and alive:
                # if it is alive, it is definitely at least starting up.
                state = InjectorState.STARTING

            if state in (InjectorState.STARTING, InjectorState.RUNNING) and not alive:
                # we thought it is running (maybe it was when get_state was previously),
                # but the process is not alive. It probably crashed
                state = InjectorState.ERROR
                logger.error("Injector was unexpectedly found stopped")

        logger.debug(
            'Injector state of "%s", "%s": %s',
            self.group.key,
            self.preset.name,
            state,
        )
        self._state = state
        return self._state

    @ensure_numlock
    def stop_injecting(self) -> None:
        """Stop injecting keycodes.

        Can be safely called from the main procss.
        """
        logger.info('Stopping injecting keycodes for group "%s"', self.group.key)
        self._msg_pipe[1].send(InjectorCommand.CLOSE)

    """Process internal stuff."""

    def _find_input_device(
        self, input_config: InputConfig
    ) -> Optional[evdev.InputDevice]:
        """find the InputDevice specified by the InputConfig

        ensures the devices supports the type and code specified by the InputConfig"""
        devices_by_hash = {get_device_hash(device): device for device in self._devices}

        # mypy thinks None is the wrong type for dict.get()
        if device := devices_by_hash.get(input_config.origin_hash):  # type: ignore
            if input_config.code in device.capabilities(absinfo=False).get(
                input_config.type, []
            ):
                return device
        return None

    def _find_input_device_fallback(
        self, input_config: InputConfig
    ) -> Optional[evdev.InputDevice]:
        """find the InputDevice specified by the InputConfig fallback logic"""
        ranking = [
            DeviceType.KEYBOARD,
            DeviceType.GAMEPAD,
            DeviceType.MOUSE,
            DeviceType.TOUCHPAD,
            DeviceType.GRAPHICS_TABLET,
            DeviceType.CAMERA,
            DeviceType.UNKNOWN,
        ]
        candidates: List[evdev.InputDevice] = [
            device
            for device in self._devices
            if input_config.code
            in device.capabilities(absinfo=False).get(input_config.type, [])
        ]

        if len(candidates) > 1:
            # there is more than on input device which can be used for this
            # event we choose only one determined by the ranking
            return sorted(candidates, key=lambda d: ranking.index(classify(d)))[0]
        if len(candidates) == 1:
            return candidates.pop()

        logger.error(f"Could not find input for {input_config}")
        return None

    def _grab_devices(self) -> Dict[DeviceHash, evdev.InputDevice]:
        """Grab all InputDevices that match a mappings' origin_hash."""
        # use a dict because the InputDevice is not directly hashable
        needed_devices = {}
        input_configs = set()

        # find all unique input_config's
        for mapping in self.preset:
            for input_config in mapping.input_combination:
                input_configs.add(input_config)

        # find all unique input_device's
        for input_config in input_configs:
            if not (device := self._find_input_device(input_config)):
                # there is no point in trying the fallback because
                # self._update_preset already did that.
                continue
            needed_devices[device.path] = device

        grabbed_devices = {}
        for device in needed_devices.values():
            if device := self._grab_device(device):
                grabbed_devices[get_device_hash(device)] = device

        return grabbed_devices

    def _update_preset(self):
        """Update all InputConfigs in the preset to include correct origin_hash
        information."""
        mappings_by_input = defaultdict(list)
        for mapping in self.preset:
            for input_config in mapping.input_combination:
                mappings_by_input[input_config].append(mapping)

        for input_config in mappings_by_input:
            if self._find_input_device(input_config):
                continue

            if not (device := self._find_input_device_fallback(input_config)):
                # fallback failed, this mapping will be ignored
                continue

            for mapping in mappings_by_input[input_config]:
                combination: List[InputConfig] = list(mapping.input_combination)
                device_hash = get_device_hash(device)
                idx = combination.index(input_config)
                combination[idx] = combination[idx].modify(origin_hash=device_hash)
                mapping.input_combination = combination

    def _grab_device(self, device: evdev.InputDevice) -> Optional[evdev.InputDevice]:
        """Try to grab the device, return None if not possible.

        Without grab, original events from it would reach the display server
        even though they are mapped.
        """
        error = None
        for attempt in range(10):
            try:
                device.grab()
                logger.debug("Grab %s", device.path)
                return device
            except IOError as err:
                # it might take a little time until the device is free if
                # it was previously grabbed.
                error = err
                logger.debug("Failed attempts to grab %s: %d", device.path, attempt + 1)
                time.sleep(self.regrab_timeout)

        logger.error("Cannot grab %s, it is possibly in use", device.path)
        logger.error(str(error))
        return None

    @staticmethod
    def _copy_capabilities(input_device: evdev.InputDevice) -> CapabilitiesDict:
        """Copy capabilities for a new device."""
        ecodes = evdev.ecodes

        # copy the capabilities because the uinput is going
        # to act like the device.
        capabilities = input_device.capabilities(absinfo=True)

        # just like what python-evdev does in from_device
        if ecodes.EV_SYN in capabilities:
            del capabilities[ecodes.EV_SYN]
        if ecodes.EV_FF in capabilities:
            del capabilities[ecodes.EV_FF]

        if ecodes.ABS_VOLUME in capabilities.get(ecodes.EV_ABS, []):
            # For some reason an ABS_VOLUME capability likes to appear
            # for some users. It prevents mice from moving around and
            # keyboards from writing symbols
            capabilities[ecodes.EV_ABS].remove(ecodes.ABS_VOLUME)

        return capabilities

    async def _msg_listener(self) -> None:
        """Wait for messages from the main process to do special stuff."""
        loop = asyncio.get_event_loop()
        while True:
            frame_available = asyncio.Event()
            loop.add_reader(self._msg_pipe[0].fileno(), frame_available.set)
            await frame_available.wait()
            frame_available.clear()
            msg = self._msg_pipe[0].recv()

            if msg == InjectorCommand.CLOSE:
                await self._close()
                return

    async def _close(self):
        logger.debug("Received close signal")
        self._stop_event.set()
        # give the event pipeline some time to reset devices
        # before shutting the loop down
        await asyncio.sleep(0.1)

        # stop the event loop and cause the process to reach its end
        # cleanly. Using .terminate prevents coverage from working.
        loop = asyncio.get_event_loop()
        loop.stop()

        self._msg_pipe[0].send(InjectorState.STOPPED)

    def _create_forwarding_device(self, source: evdev.InputDevice) -> evdev.UInput:
        # copy as much information as possible, because libinput uses the extra
        # information to enable certain features like "Disable touchpad while
        # typing"
        try:
            forward_to = evdev.UInput(
                name=get_udev_name(source.name, "forwarded"),
                events=self._copy_capabilities(source),
                # phys=source.phys,  # this leads to confusion. the appearance of
                # a uinput with this "phys" property causes the udev rule to
                # autoload for the original device, overwriting our previous
                # attempts at starting an injection.
                vendor=source.info.vendor,
                product=source.info.product,
                version=source.info.version,
                bustype=source.info.bustype,
                input_props=source.input_props(),
            )
        except TypeError as e:
            if "input_props" in str(e):
                # UInput constructor doesn't support input_props and
                # source.input_props doesn't exist with old python-evdev versions.
                logger.error("Please upgrade your python-evdev version. Exiting")
                self._msg_pipe[0].send(InjectorState.UPGRADE_EVDEV)
                sys.exit(12)

            raise e
        return forward_to

    def run(self) -> None:
        """The injection worker that keeps injecting until terminated.

        Stuff is non-blocking by using asyncio in order to do multiple things
        somewhat concurrently.

        Use this function as starting point in a process. It creates
        the loops needed to read and map events and keeps running them.
        """
        logger.info('Starting injecting the preset for "%s"', self.group.key)

        # create a new event loop, because somehow running an infinite loop
        # that sleeps on iterations (joystick_to_mouse) in one process causes
        # another injection process to screw up reading from the grabbed
        # device.
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        self._devices = self.group.get_devices()

        # InputConfigs may not contain the origin_hash information, this will try to
        # make a good guess if the origin_hash information is missing or invalid.
        self._update_preset()

        # grab devices as early as possible. If events appear that won't get
        # released anymore before the grab they appear to be held down forever
        sources = self._grab_devices()
        forward_devices = {}
        for device_hash, device in sources.items():
            forward_devices[device_hash] = self._create_forwarding_device(device)

        # create this within the process after the event loop creation,
        # so that the macros use the correct loop
        self.context = Context(
            self.preset,
            sources,
            forward_devices,
            self.mapping_parser,
        )
        self._stop_event = asyncio.Event()

        if len(sources) == 0:
            # maybe the preset was empty or something
            logger.error("Did not grab any device")
            self._msg_pipe[0].send(InjectorState.NO_GRAB)
            return

        numlock_state = is_numlock_on()
        coroutines = []

        for device_hash in sources:
            # actually doing things
            event_reader = EventReader(
                self.context,
                sources[device_hash],
                self._stop_event,
            )
            coroutines.append(event_reader.run())
            self._event_readers.append(event_reader)

        coroutines.append(self._msg_listener())

        # set the numlock state to what it was before injecting, because
        # grabbing devices screws this up
        set_numlock(numlock_state)

        self._msg_pipe[0].send(InjectorState.RUNNING)

        try:
            loop.run_until_complete(asyncio.gather(*coroutines))
        except RuntimeError as error:
            # the loop might have been stopped via a `CLOSE` message,
            # which causes the error message below. This is expected behavior
            if str(error) != "Event loop stopped before Future completed.":
                raise error
        except OSError as error:
            logger.error("Failed to run injector coroutines: %s", str(error))

        if len(coroutines) > 0:
            # expected when stop_injecting is called,
            # during normal operation as well as tests this point is not
            # reached otherwise.
            logger.debug("Injector coroutines ended")

        for source in sources.values():
            # ungrab at the end to make the next injection process not fail
            # its grabs
            try:
                source.ungrab()
            except OSError as error:
                # it might have disappeared
                logger.debug("OSError for ungrab on %s: %s", source.path, str(error))



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import Tuple, Optional, Hashable, Literal

import evdev
from evdev import ecodes

from inputremapper.utils import get_evdev_constant_name, DeviceHash


class EventActions(enum.Enum):
    """Additional information an InputEvent can send through the event pipeline."""

    as_key = enum.auto()  # treat this event as a key event
    recenter = enum.auto()  # recenter the axis when receiving this
    none = enum.auto()

    # used in combination with as_key, for originally abs or rel events
    positive_trigger = enum.auto()  # original event was positive direction
    negative_trigger = enum.auto()  # original event was negative direction


# Todo: add slots=True as soon as python 3.10 is in common distros
@dataclass(frozen=True)
class InputEvent:
    """Events that are generated during runtime.

    Is a drop-in replacement for evdev.InputEvent
    """

    sec: int
    usec: int
    type: int
    code: int
    value: int
    actions: Tuple[EventActions, ...] = ()
    origin_hash: Optional[DeviceHash] = None

    def __eq__(self, other: InputEvent | evdev.InputEvent | Tuple[int, int, int]):
        # useful in tests
        if isinstance(other, InputEvent) or isinstance(other, evdev.InputEvent):
            return self.event_tuple == (other.type, other.code, other.value)
        if isinstance(other, tuple):
            return self.event_tuple == other
        raise TypeError(f"cannot compare {type(other)} with InputEvent")

    @staticmethod
    def validate_event(event):
        """Test if the event is valid."""
        if not isinstance(event.type, int):
            raise TypeError(f"Expected type to be an int, but got {event.type}")

        if not isinstance(event.code, int):
            raise TypeError(f"Expected code to be an int, but got {event.code}")

        if not isinstance(event.value, int):
            # this happened to me because I screwed stuff up
            raise TypeError(f"Expected value to be an int, but got {event.value}")

        return event

    @property
    def input_match_hash(self) -> Hashable:
        """a Hashable object which is intended to match the InputEvent with a
        InputConfig.
        """
        return self.type, self.code, self.origin_hash

    @classmethod
    def from_event(
        cls,
        event: evdev.InputEvent,
        origin_hash: Optional[DeviceHash] = None,
    ) -> InputEvent:
        """Create a InputEvent from another InputEvent or evdev.InputEvent."""
        try:
            return cls(
                event.sec,
                event.usec,
                event.type,
                event.code,
                event.value,
                origin_hash=origin_hash,
            )
        except AttributeError as exception:
            raise TypeError(
                f"Failed to create InputEvent from {event = }"
            ) from exception

    @classmethod
    def from_tuple(
        cls,
        event_tuple: Tuple[int, int, int],
        origin_hash: Optional[DeviceHash] = None,
    ) -> InputEvent:
        """Create a InputEvent from a (type, code, value) tuple."""
        # use this as rarely as possible. Construct objects early on and pass them
        # around instead of passing around integers
        if len(event_tuple) != 3:
            raise TypeError(
                f"failed to create InputEvent {event_tuple = } must have length 3"
            )

        return cls.validate_event(
            cls(
                0,
                0,
                int(event_tuple[0]),
                int(event_tuple[1]),
                int(event_tuple[2]),
                origin_hash=origin_hash,
            )
        )

    @classmethod
    def abs(cls, code: int, value: int, origin_hash: Optional[DeviceHash] = None):
        """Create an abs event, like joystick movements."""
        return cls.validate_event(
            cls(
                0,
                0,
                ecodes.EV_ABS,
                code,
                value,
                origin_hash=origin_hash,
            )
        )

    @classmethod
    def rel(cls, code: int, value: int, origin_hash: Optional[str] = None):
        """Create a rel event, like mouse movements."""
        return cls.validate_event(
            cls(
                0,
                0,
                ecodes.EV_REL,
                code,
                value,
                origin_hash=origin_hash,
            )
        )

    @classmethod
    def key(cls, code: int, value: Literal[0, 1], origin_hash: Optional[str] = None):
        """Create a key event, like keyboard keys or gamepad buttons.

        A value of 1 means "press", a value of 0 means "release".
        """
        return cls.validate_event(
            cls(
                0,
                0,
                ecodes.EV_KEY,
                code,
                value,
                origin_hash=origin_hash,
            )
        )

    @property
    def type_and_code(self) -> Tuple[int, int]:
        """Event type, code."""
        return self.type, self.code

    @property
    def event_tuple(self) -> Tuple[int, int, int]:
        """Event type, code, value."""
        return self.type, self.code, self.value

    @property
    def is_key_event(self) -> bool:
        """Whether this is interpreted as a key event."""
        return self.type == evdev.ecodes.EV_KEY or EventActions.as_key in self.actions

    @property
    def is_wheel_event(self) -> bool:
        """Whether this is interpreted as a key event."""
        return self.type == evdev.ecodes.EV_REL and self.code in [
            ecodes.REL_WHEEL,
            ecodes.REL_HWHEEL,
        ]

    @property
    def is_wheel_hi_res_event(self) -> bool:
        """Whether this is interpreted as a key event."""
        return self.type == evdev.ecodes.EV_REL and self.code in [
            ecodes.REL_WHEEL_HI_RES,
            ecodes.REL_HWHEEL_HI_RES,
        ]

    def __str__(self):
        name = get_evdev_constant_name(self.type, self.code)
        return f"InputEvent for {self.event_tuple} {name}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    def timestamp(self):
        """Return the unix timestamp of when the event was seen."""
        return self.sec + self.usec / 1000000

    def modify(
        self,
        sec: Optional[int] = None,
        usec: Optional[int] = None,
        type_: Optional[int] = None,
        code: Optional[int] = None,
        value: Optional[int] = None,
        actions: Optional[Tuple[EventActions, ...]] = None,
        origin_hash: Optional[str] = None,
    ) -> InputEvent:
        """Return a new modified event."""
        return InputEvent(
            sec if sec is not None else self.sec,
            usec if usec is not None else self.usec,
            type_ if type_ is not None else self.type,
            code if code is not None else self.code,
            value if value is not None else self.value,
            actions if actions is not None else self.actions,
            origin_hash=origin_hash if origin_hash is not None else self.origin_hash,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.task import Task
from inputremapper.logging.logger import logger


class AddTask(Task):
    """Add a number to a variable."""

    argument_configs = [
        ArgumentConfig(
            name="variable",
            position=0,
            types=[int, float, None],
            is_variable_name=True,
        ),
        ArgumentConfig(
            name="value",
            position=1,
            types=[int, float],
        ),
    ]

    async def run(self, callback) -> None:
        argument = self.get_argument("variable")
        try:
            current = argument.get_value()
        except MacroError:
            return

        if current is None:
            logger.debug(
                '"%s" initialized with 0',
                self.arguments["variable"].get_variable_name(),
            )
            argument.set_value(0)
            current = 0

        addend = self.get_argument("value").get_value()

        if not isinstance(current, (int, float)):
            logger.error(
                'Expected variable "%s" to contain a number, but got "%s"',
                argument.get_value(),
                current,
            )
            return

        logger.debug("%s += %s", current, addend)
        argument.set_value(current + addend)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

xmodmap = (
    b"keycode   8 =\nkeycode   9 = Escape NoSymbol Escape\nkeycode  10 = 1 exclam 1 exclam onesuperior exclamdown ones"
    b"uperior\nkeycode  11 = 2 quotedbl 2 quotedbl twosuperior oneeighth twosuperior\nkeycode  12 = 3 section 3 sectio"
    b"n threesuperior sterling threesuperior\nkeycode  13 = 4 dollar 4 dollar onequarter currency onequarter\nkeycode "
    b" 14 = 5 percent 5 percent onehalf threeeighths onehalf\nkeycode  15 = 6 ampersand 6 ampersand notsign fiveeighth"
    b"s notsign\nkeycode  16 = 7 slash 7 slash braceleft seveneighths braceleft\nkeycode  17 = 8 parenleft 8 parenleft"
    b" bracketleft trademark bracketleft\nkeycode  18 = 9 parenright 9 parenright bracketright plusminus bracketright"
    b"\nkeycode  19 = 0 equal 0 equal braceright degree braceright\nkeycode  20 = ssharp question ssharp question back"
    b"slash questiondown U1E9E\nkeycode  21 = dead_acute dead_grave dead_acute dead_grave dead_cedilla dead_ogonek dea"
    b"d_cedilla\nkeycode  22 = BackSpace BackSpace BackSpace BackSpace\nkeycode  23 = Tab ISO_Left_Tab Tab ISO_Left_Ta"
    b"b\nkeycode  24 = q Q q Q at Greek_OMEGA at\nkeycode  25 = w W w W lstroke Lstroke lstroke\nkeycode  26 = e E e E"
    b" EuroSign EuroSign EuroSign\nkeycode  27 = r R r R paragraph registered paragraph\nkeycode  28 = t T t T tslash "
    b"Tslash tslash\nkeycode  29 = z Z z Z leftarrow yen leftarrow\nkeycode  30 = u U u U downarrow uparrow downarrow"
    b"\nkeycode  31 = i I i I rightarrow idotless rightarrow\nkeycode  32 = o O o O oslash Oslash oslash\nkeycode  33 "
    b"= p P p P thorn THORN thorn\nkeycode  34 = udiaeresis Udiaeresis udiaeresis Udiaeresis dead_diaeresis dead_above"
    b"ring dead_diaeresis\nkeycode  35 = plus asterisk plus asterisk asciitilde macron asciitilde\nkeycode  36 = Retur"
    b"n NoSymbol Return\nkeycode  37 = Control_L NoSymbol Control_L\nkeycode  38 = a A a A ae AE ae\nkeycode  39 = s S"
    b" s S U017F U1E9E U017F\nkeycode  40 = d D d D eth ETH eth\nkeycode  41 = f F f F dstroke ordfeminine dstroke\nke"
    b"ycode  42 = g G g G eng ENG eng\nkeycode  43 = h H h H hstroke Hstroke hstroke\nkeycode  44 = j J j J dead_below"
    b"dot dead_abovedot dead_belowdot\nkeycode  45 = k K k K kra ampersand kra\nkeycode  46 = l L l L lstroke Lstroke "
    b"lstroke\nkeycode  47 = odiaeresis Odiaeresis odiaeresis Odiaeresis dead_doubleacute dead_belowdot dead_doubleacu"
    b"te\nkeycode  48 = adiaeresis Adiaeresis adiaeresis Adiaeresis dead_circumflex dead_caron dead_circumflex\nkeycod"
    b"e  49 = dead_circumflex degree dead_circumflex degree U2032 U2033 U2032\nkeycode  50 = Shift_L NoSymbol Shift_L"
    b"\nkeycode  51 = numbersign apostrophe numbersign apostrophe rightsinglequotemark dead_breve rightsinglequotemark"
    b"\nkeycode  52 = y Y y Y guillemotright U203A guillemotright\nkeycode  53 = x X x X guillemotleft U2039 guillemot"
    b"left\nkeycode  54 = c C c C cent copyright cent\nkeycode  55 = v V v V doublelowquotemark singlelowquotemark dou"
    b"blelowquotemark\nkeycode  56 = b B b B leftdoublequotemark leftsinglequotemark leftdoublequotemark\nkeycode  57 "
    b"= n N n N rightdoublequotemark rightsinglequotemark rightdoublequotemark\nkeycode  58 = m M m M mu masculine mu"
    b"\nkeycode  59 = comma semicolon comma semicolon periodcentered multiply periodcentered\nkeycode  60 = period col"
    b"on period colon U2026 division U2026\nkeycode  61 = minus underscore minus underscore endash emdash endash\nkeyc"
    b"ode  62 = Shift_R NoSymbol Shift_R\nkeycode  63 = KP_Multiply KP_Multiply KP_Multiply KP_Multiply KP_Multiply KP"
    b"_Multiply XF86ClearGrab\nkeycode  64 = Alt_L Meta_L Alt_L Meta_L\nkeycode  65 = space NoSymbol space\nkeycode  6"
    b"6 = Caps_Lock NoSymbol Caps_Lock\nkeycode  67 = F1 F1 F1 F1 F1 F1 XF86Switch_VT_1\nkeycode  68 = F2 F2 F2 F2 F2 "
    b"F2 XF86Switch_VT_2\nkeycode  69 = F3 F3 F3 F3 F3 F3 XF86Switch_VT_3\nkeycode  70 = F4 F4 F4 F4 F4 F4 XF86Switch_"
    b"VT_4\nkeycode  71 = F5 F5 F5 F5 F5 F5 XF86Switch_VT_5\nkeycode  72 = F6 F6 F6 F6 F6 F6 XF86Switch_VT_6\nkeycode "
    b" 73 = F7 F7 F7 F7 F7 F7 XF86Switch_VT_7\nkeycode  74 = F8 F8 F8 F8 F8 F8 XF86Switch_VT_8\nkeycode  75 = F9 F9 F9"
    b" F9 F9 F9 XF86Switch_VT_9\nkeycode  76 = F10 F10 F10 F10 F10 F10 XF86Switch_VT_10\nkeycode  77 = Num_Lock NoSymb"
    b"ol Num_Lock\nkeycode  78 = Scroll_Lock NoSymbol Scroll_Lock\nkeycode  79 = KP_Home KP_7 KP_Home KP_7\nkeycode  8"
    b"0 = KP_Up KP_8 KP_Up KP_8\nkeycode  81 = KP_Prior KP_9 KP_Prior KP_9\nkeycode  82 = KP_Subtract KP_Subtract KP_S"
    b"ubtract KP_Subtract KP_Subtract KP_Subtract XF86Prev_VMode\nkeycode  83 = KP_Left KP_4 KP_Left KP_4\nkeycode  84"
    b" = KP_Begin KP_5 KP_Begin KP_5\nkeycode  85 = KP_Right KP_6 KP_Right KP_6\nkeycode  86 = KP_Add KP_Add KP_Add KP"
    b"_Add KP_Add KP_Add XF86Next_VMode\nkeycode  87 = KP_End KP_1 KP_End KP_1\nkeycode  88 = KP_Down KP_2 KP_Down KP_"
    b"2\nkeycode  89 = KP_Next KP_3 KP_Next KP_3\nkeycode  90 = KP_Insert KP_0 KP_Insert KP_0\nkeycode  91 = KP_Delete"
    b" KP_Separator KP_Delete KP_Separator\nkeycode  92 = ISO_Level3_Shift NoSymbol ISO_Level3_Shift\nkeycode  93 =\nk"
    b"eycode  94 = less greater less greater bar dead_belowmacron bar\nkeycode  95 = F11 F11 F11 F11 F11 F11 XF86Switc"
    b"h_VT_11\nkeycode  96 = F12 F12 F12 F12 F12 F12 XF86Switch_VT_12\nkeycode  97 =\nkeycode  98 = Katakana NoSymbol "
    b"Katakana\nkeycode  99 = Hiragana NoSymbol Hiragana\nkeycode 100 = Henkan_Mode NoSymbol Henkan_Mode\nkeycode 101 "
    b"= Hiragana_Katakana NoSymbol Hiragana_Katakana\nkeycode 102 = Muhenkan NoSymbol Muhenkan\nkeycode 103 =\nkeycode"
    b" 104 = KP_Enter NoSymbol KP_Enter\nkeycode 105 = Control_R NoSymbol Control_R\nkeycode 106 = KP_Divide KP_Divide"
    b" KP_Divide KP_Divide KP_Divide KP_Divide XF86Ungrab\nkeycode 107 = Print Sys_Req Print Sys_Req\nkeycode 108 = IS"
    b"O_Level3_Shift NoSymbol ISO_Level3_Shift\nkeycode 109 = Linefeed NoSymbol Linefeed\nkeycode 110 = Home NoSymbol "
    b"Home\nkeycode 111 = Up NoSymbol Up\nkeycode 112 = Prior NoSymbol Prior\nkeycode 113 = Left NoSymbol Left\nkeycod"
    b"e 114 = Right NoSymbol Right\nkeycode 115 = End NoSymbol End\nkeycode 116 = Down NoSymbol Down\nkeycode 117 = Ne"
    b"xt NoSymbol Next\nkeycode 118 = Insert NoSymbol Insert\nkeycode 119 = Delete NoSymbol Delete\nkeycode 120 =\nkey"
    b"code 121 = XF86AudioMute NoSymbol XF86AudioMute\nkeycode 122 = XF86AudioLowerVolume NoSymbol XF86AudioLowerVolum"
    b"e\nkeycode 123 = XF86AudioRaiseVolume NoSymbol XF86AudioRaiseVolume\nkeycode 124 = XF86PowerOff NoSymbol XF86Pow"
    b"erOff\nkeycode 125 = KP_Equal NoSymbol KP_Equal\nkeycode 126 = plusminus NoSymbol plusminus\nkeycode 127 = Pause"
    b" Break Pause Break\nkeycode 128 = XF86LaunchA NoSymbol XF86LaunchA\nkeycode 129 = KP_Decimal KP_Decimal KP_Decim"
    b"al KP_Decimal\nkeycode 130 = Hangul NoSymbol Hangul\nkeycode 131 = Hangul_Hanja NoSymbol Hangul_Hanja\nkeycode 1"
    b"32 =\nkeycode 133 = Super_L NoSymbol Super_L\nkeycode 134 = Super_R NoSymbol Super_R\nkeycode 135 = Menu NoSymbo"
    b"l Menu\nkeycode 136 = Cancel NoSymbol Cancel\nkeycode 137 = Redo NoSymbol Redo\nkeycode 138 = SunProps NoSymbol "
    b"SunProps\nkeycode 139 = Undo NoSymbol Undo\nkeycode 140 = SunFront NoSymbol SunFront\nkeycode 141 = XF86Copy NoS"
    b"ymbol XF86Copy\nkeycode 142 = XF86Open NoSymbol XF86Open\nkeycode 143 = XF86Paste NoSymbol XF86Paste\nkeycode 14"
    b"4 = Find NoSymbol Find\nkeycode 145 = XF86Cut NoSymbol XF86Cut\nkeycode 146 = Help NoSymbol Help\nkeycode 147 = "
    b"XF86MenuKB NoSymbol XF86MenuKB\nkeycode 148 = XF86Calculator NoSymbol XF86Calculator\nkeycode 149 =\nkeycode 150"
    b" = XF86Sleep NoSymbol XF86Sleep\nkeycode 151 = XF86WakeUp NoSymbol XF86WakeUp\nkeycode 152 = XF86Explorer NoSymb"
    b"ol XF86Explorer\nkeycode 153 = XF86Send NoSymbol XF86Send\nkeycode 154 =\nkeycode 155 = XF86Xfer NoSymbol XF86Xf"
    b"er\nkeycode 156 = XF86Launch1 NoSymbol XF86Launch1\nkeycode 157 = XF86Launch2 NoSymbol XF86Launch2\nkeycode 158 "
    b"= XF86WWW NoSymbol XF86WWW\nkeycode 159 = XF86DOS NoSymbol XF86DOS\nkeycode 160 = XF86ScreenSaver NoSymbol XF86S"
    b"creenSaver\nkeycode 161 = XF86RotateWindows NoSymbol XF86RotateWindows\nkeycode 162 = XF86TaskPane NoSymbol XF86"
    b"TaskPane\nkeycode 163 = XF86Mail NoSymbol XF86Mail\nkeycode 164 = XF86Favorites NoSymbol XF86Favorites\nkeycode "
    b"165 = XF86MyComputer NoSymbol XF86MyComputer\nkeycode 166 = XF86Back NoSymbol XF86Back\nkeycode 167 = XF86Forwar"
    b"d NoSymbol XF86Forward\nkeycode 168 =\nkeycode 169 = XF86Eject NoSymbol XF86Eject\nkeycode 170 = XF86Eject XF86E"
    b"ject XF86Eject XF86Eject\nkeycode 171 = XF86AudioNext NoSymbol XF86AudioNext\nkeycode 172 = XF86AudioPlay XF86Au"
    b"dioPause XF86AudioPlay XF86AudioPause\nkeycode 173 = XF86AudioPrev NoSymbol XF86AudioPrev\nkeycode 174 = XF86Aud"
    b"ioStop XF86Eject XF86AudioStop XF86Eject\nkeycode 175 = XF86AudioRecord NoSymbol XF86AudioRecord\nkeycode 176 = "
    b"XF86AudioRewind NoSymbol XF86AudioRewind\nkeycode 177 = XF86Phone NoSymbol XF86Phone\nkeycode 178 =\nkeycode 179"
    b" = XF86Tools NoSymbol XF86Tools\nkeycode 180 = XF86HomePage NoSymbol XF86HomePage\nkeycode 181 = XF86Reload NoSy"
    b"mbol XF86Reload\nkeycode 182 = XF86Close NoSymbol XF86Close\nkeycode 183 =\nkeycode 184 =\nkeycode 185 = XF86Scr"
    b"ollUp NoSymbol XF86ScrollUp\nkeycode 186 = XF86ScrollDown NoSymbol XF86ScrollDown\nkeycode 187 = parenleft NoSym"
    b"bol parenleft\nkeycode 188 = parenright NoSymbol parenright\nkeycode 189 = XF86New NoSymbol XF86New\nkeycode 190"
    b" = Redo NoSymbol Redo\nkeycode 191 = XF86Tools NoSymbol XF86Tools\nkeycode 192 = XF86Launch5 NoSymbol XF86Launch"
    b"5\nkeycode 193 = XF86Launch6 NoSymbol XF86Launch6\nkeycode 194 = XF86Launch7 NoSymbol XF86Launch7\nkeycode 195 ="
    b" XF86Launch8 NoSymbol XF86Launch8\nkeycode 196 = XF86Launch9 NoSymbol XF86Launch9\nkeycode 197 =\nkeycode 198 = "
    b"XF86AudioMicMute NoSymbol XF86AudioMicMute\nkeycode 199 = XF86TouchpadToggle NoSymbol XF86TouchpadToggle\nkeycod"
    b"e 200 = XF86TouchpadOn NoSymbol XF86TouchpadOn\nkeycode 201 = XF86TouchpadOff NoSymbol XF86TouchpadOff\nkeycode "
    b"202 =\nkeycode 203 = Mode_switch NoSymbol Mode_switch\nkeycode 204 = NoSymbol Alt_L NoSymbol Alt_L\nkeycode 205 "
    b"= NoSymbol Meta_L NoSymbol Meta_L\nkeycode 206 = NoSymbol Super_L NoSymbol Super_L\nkeycode 207 = NoSymbol Hyper"
    b"_L NoSymbol Hyper_L\nkeycode 208 = XF86AudioPlay NoSymbol XF86AudioPlay\nkeycode 209 = XF86AudioPause NoSymbol X"
    b"F86AudioPause\nkeycode 210 = XF86Launch3 NoSymbol XF86Launch3\nkeycode 211 = XF86Launch4 NoSymbol XF86Launch4\nk"
    b"eycode 212 = XF86LaunchB NoSymbol XF86LaunchB\nkeycode 213 = XF86Suspend NoSymbol XF86Suspend\nkeycode 214 = XF8"
    b"6Close NoSymbol XF86Close\nkeycode 215 = XF86AudioPlay NoSymbol XF86AudioPlay\nkeycode 216 = XF86AudioForward No"
    b"Symbol XF86AudioForward\nkeycode 217 =\nkeycode 218 = Print NoSymbol Print\nkeycode 219 =\nkeycode 220 = XF86Web"
    b"Cam NoSymbol XF86WebCam\nkeycode 221 = XF86AudioPreset NoSymbol XF86AudioPreset\nkeycode 222 =\nkeycode 223 = XF"
    b"86Mail NoSymbol XF86Mail\nkeycode 224 = XF86Messenger NoSymbol XF86Messenger\nkeycode 225 = XF86Search NoSymbol "
    b"XF86Search\nkeycode 226 = XF86Go NoSymbol XF86Go\nkeycode 227 = XF86Finance NoSymbol XF86Finance\nkeycode 228 = "
    b"XF86Game NoSymbol XF86Game\nkeycode 229 = XF86Shop NoSymbol XF86Shop\nkeycode 230 =\nkeycode 231 = Cancel NoSymb"
    b"ol Cancel\nkeycode 232 = XF86MonBrightnessDown NoSymbol XF86MonBrightnessDown\nkeycode 233 = XF86MonBrightnessUp"
    b" NoSymbol XF86MonBrightnessUp\nkeycode 234 = XF86AudioMedia NoSymbol XF86AudioMedia\nkeycode 235 = XF86Display N"
    b"oSymbol XF86Display\nkeycode 236 = XF86KbdLightOnOff NoSymbol XF86KbdLightOnOff\nkeycode 237 = XF86KbdBrightness"
    b"Down NoSymbol XF86KbdBrightnessDown\nkeycode 238 = XF86KbdBrightnessUp NoSymbol XF86KbdBrightnessUp\nkeycode 239"
    b" = XF86Send NoSymbol XF86Send\nkeycode 240 = XF86Reply NoSymbol XF86Reply\nkeycode 241 = XF86MailForward NoSymbo"
    b"l XF86MailForward\nkeycode 242 = XF86Save NoSymbol XF86Save\nkeycode 243 = XF86Documents NoSymbol XF86Documents"
    b"\nkeycode 244 = XF86Battery NoSymbol XF86Battery\nkeycode 245 = XF86Bluetooth NoSymbol XF86Bluetooth\nkeycode 24"
    b"6 = XF86WLAN NoSymbol XF86WLAN\nkeycode 247 =\nkeycode 248 =\nkeycode 249 =\nkeycode 250 =\nkeycode 251 = XF86Mo"
    b"nBrightnessCycle NoSymbol XF86MonBrightnessCycle\nkeycode 252 =\nkeycode 253 =\nkeycode 254 = XF86WWAN NoSymbol "
    b"XF86WWAN\nkeycode 255 = XF86RFKill NoSymbol XF86RFKill\n"
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestToggle(MacroTestBase):
    async def test_toggle(self):
        # repeats key(a) as long as macro is toggled
        macro = Parser.parse("toggle(key(a))", self.context, DummyMapping)
        code_a = keyboard_layout.get("a")

        self.assertEqual(self.count_child_macros(macro), 1)
        self.assertEqual(self.count_tasks(macro), 2)

        # Start
        macro.press_trigger()
        macro.release_trigger()
        asyncio.ensure_future(macro.run(self.handler))

        await asyncio.sleep(0.1)
        count_1 = self.result.count((EV_KEY, code_a, 1))
        self.assertGreater(count_1, 2)

        # Keeps writing more
        await asyncio.sleep(0.1)
        count_2 = self.result.count((EV_KEY, code_a, 1))
        self.assertGreater(count_2, count_1)

        # Stop
        macro.press_trigger()
        macro.release_trigger()

        count_3 = self.result.count((EV_KEY, code_a, 1))
        await asyncio.sleep(0.1)
        count_4 = self.result.count((EV_KEY, code_a, 1))
        # ensure that the macro has stopped
        self.assertEqual(count_3, count_4)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import multiprocessing
import os
import time
import unittest

import evdev
from evdev.ecodes import EV_ABS, EV_KEY

from inputremapper.groups import groups, _Groups
from inputremapper.gui.messages.message_broker import MessageBroker
from inputremapper.gui.reader_client import ReaderClient
from inputremapper.gui.reader_service import ReaderService
from inputremapper.injection.global_uinputs import UInput, GlobalUInputs
from inputremapper.input_event import InputEvent
from inputremapper.utils import get_device_hash
from tests.lib.cleanup import cleanup
from tests.lib.constants import EVENT_READ_TIMEOUT, START_READING_DELAY
from tests.lib.fixtures import fixtures
from tests.lib.logger import logger
from tests.lib.patches import InputDevice
from tests.lib.pipes import push_events
from tests.lib.test_setup import test_setup


@test_setup
class TestTest(unittest.TestCase):
    def test_stubs(self):
        self.assertIsNotNone(groups.find(key="Foo Device 2"))

    def test_fake_capabilities(self):
        device = InputDevice("/dev/input/event30")
        capabilities = device.capabilities(absinfo=False)
        self.assertIsInstance(capabilities, dict)
        self.assertIsInstance(capabilities[EV_ABS], list)
        self.assertIsInstance(capabilities[EV_ABS][0], int)

        capabilities = device.capabilities()
        self.assertIsInstance(capabilities, dict)
        self.assertIsInstance(capabilities[EV_ABS], list)
        self.assertIsInstance(capabilities[EV_ABS][0], tuple)
        self.assertIsInstance(capabilities[EV_ABS][0][0], int)
        self.assertIsInstance(capabilities[EV_ABS][0][1], evdev.AbsInfo)
        self.assertIsInstance(capabilities[EV_ABS][0][1].max, int)
        self.assertIsInstance(capabilities, dict)
        self.assertIsInstance(capabilities[EV_KEY], list)
        self.assertIsInstance(capabilities[EV_KEY][0], int)

    def test_restore_fixtures(self):
        fixtures["/bar/dev"] = {"name": "bla"}
        cleanup()
        self.assertIsNone(fixtures.get("/bar/dev"))
        self.assertIsNotNone(fixtures.get("/dev/input/event11"))

    def test_restore_os_environ(self):
        os.environ["foo"] = "bar"
        del os.environ["USER"]
        environ = os.environ
        cleanup()
        self.assertIn("USER", environ)
        self.assertNotIn("foo", environ)

    def test_push_events(self):
        """Test that push_event works properly between reader service and client.

        Using push_events after the reader-service is already started should work,
        as well as using push_event twice
        """
        reader_client = ReaderClient(MessageBroker(), groups)

        def create_reader_service():
            # this will cause pending events to be copied over to the reader-service
            # process
            def start_reader_service():
                # Create dependencies from scratch, because the reader-service runs
                # in a different process
                global_uinputs = GlobalUInputs(UInput)
                reader_service = ReaderService(_Groups(), global_uinputs)
                loop = asyncio.new_event_loop()
                loop.run_until_complete(reader_service.run())

            self.reader_service = multiprocessing.Process(target=start_reader_service)
            self.reader_service.start()
            time.sleep(0.1)

        def wait_for_results():
            # wait for the reader-service to send stuff
            for _ in range(10):
                time.sleep(EVENT_READ_TIMEOUT)
                if reader_client._results_pipe.poll():
                    break

        create_reader_service()
        reader_client.set_group(groups.find(key="Foo Device 2"))
        reader_client.start_recorder()
        time.sleep(START_READING_DELAY)

        event = InputEvent.key(102, 1)
        push_events(fixtures.foo_device_2_keyboard, [event])
        wait_for_results()
        self.assertTrue(reader_client._results_pipe.poll())

        reader_client._read()
        self.assertFalse(reader_client._results_pipe.poll())

        # can push more events to the reader-service that is inside a separate
        # process, which end up being sent to the reader
        event = InputEvent.key(102, 0)
        logger.info("push_events")
        push_events(fixtures.foo_device_2_keyboard, [event])
        wait_for_results()
        logger.info("assert")
        self.assertTrue(reader_client._results_pipe.poll())

        reader_client.terminate()

    def test_device_hash_from_fixture_is_correct(self):
        for fixture in fixtures:
            self.assertEqual(
                fixture.get_device_hash(), get_device_hash(InputDevice(fixture.path))
            )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Starts injecting keycodes based on the configuration."""

import sys
import multiprocessing
from argparse import ArgumentParser

from inputremapper.configs.global_config import GlobalConfig
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.logging.logger import logger


class InputRemapperServiceBin:
    @staticmethod
    def main() -> None:
        parser = ArgumentParser()
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            help="Displays additional debug information",
            default=False,
        )
        parser.add_argument(
            "--hide-info",
            action="store_true",
            dest="hide_info",
            help="Don't display version information",
            default=False,
        )

        options = parser.parse_args(sys.argv[1:])

        # Python 3.14 compatibility
        multiprocessing.set_start_method("fork")

        logger.update_verbosity(options.debug)

        # import input-remapper stuff after setting the log verbosity
        from inputremapper.daemon import Daemon

        if not options.hide_info:
            logger.log_info("input-remapper-service")

        global_config = GlobalConfig()
        global_uinputs = GlobalUInputs(UInput)
        mapping_parser = MappingParser(global_uinputs)

        daemon = Daemon(global_config, global_uinputs, mapping_parser)
        daemon.publish()
        daemon.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import subprocess
import unittest
from unittest.mock import patch

from evdev.ecodes import BTN_LEFT, KEY_A

from inputremapper.configs.paths import PathUtils
from inputremapper.configs.keyboard_layout import KeyboardLayout, XMODMAP_FILENAME
from tests.lib.test_setup import test_setup


@test_setup
class TestSystemMapping(unittest.TestCase):
    def test_update(self):
        keyboard_layout = KeyboardLayout()
        keyboard_layout.update({"foo1": 101, "bar1": 102})
        keyboard_layout.update({"foo2": 201, "bar2": 202})
        self.assertEqual(keyboard_layout.get("foo1"), 101)
        self.assertEqual(keyboard_layout.get("bar2"), 202)

    def test_xmodmap_file(self):
        keyboard_layout = KeyboardLayout()
        path = os.path.join(PathUtils.config_path(), XMODMAP_FILENAME)
        os.remove(path)

        keyboard_layout.populate()
        self.assertTrue(os.path.exists(path))
        with open(path, "r") as file:
            content = json.load(file)
            self.assertEqual(content["a"], KEY_A)
            # only xmodmap stuff should be present
            self.assertNotIn("key_a", content)
            self.assertNotIn("KEY_A", content)
            self.assertNotIn("disable", content)

    def test_empty_xmodmap(self):
        # if xmodmap returns nothing, don't write the file
        empty_xmodmap = ""

        class SubprocessMock:
            def decode(self):
                return empty_xmodmap

        def check_output(*args, **kwargs):
            return SubprocessMock()

        with patch.object(subprocess, "check_output", check_output):
            keyboard_layout = KeyboardLayout()
            path = os.path.join(PathUtils.config_path(), XMODMAP_FILENAME)
            os.remove(path)

            keyboard_layout.populate()
            self.assertFalse(os.path.exists(path))

    def test_xmodmap_command_missing(self):
        # if xmodmap is not installed, don't write the file
        def check_output(*args, **kwargs):
            raise FileNotFoundError

        with patch.object(subprocess, "check_output", check_output):
            keyboard_layout = KeyboardLayout()
            path = os.path.join(PathUtils.config_path(), XMODMAP_FILENAME)
            os.remove(path)

            keyboard_layout.populate()
            self.assertFalse(os.path.exists(path))

    def test_correct_case(self):
        keyboard_layout = KeyboardLayout()
        keyboard_layout.clear()
        keyboard_layout._set("A", 31)
        keyboard_layout._set("a", 32)
        keyboard_layout._set("abcd_B", 33)

        self.assertEqual(keyboard_layout.correct_case("a"), "a")
        self.assertEqual(keyboard_layout.correct_case("A"), "A")
        self.assertEqual(keyboard_layout.correct_case("ABCD_b"), "abcd_B")
        # unknown stuff is returned as is
        self.assertEqual(keyboard_layout.correct_case("FOo"), "FOo")

        self.assertEqual(keyboard_layout.get("A"), 31)
        self.assertEqual(keyboard_layout.get("a"), 32)
        self.assertEqual(keyboard_layout.get("ABCD_b"), 33)
        self.assertEqual(keyboard_layout.get("abcd_B"), 33)

    def test_keyboard_layout(self):
        keyboard_layout = KeyboardLayout()
        keyboard_layout.populate()
        self.assertGreater(len(keyboard_layout._mapping), 100)

        # this is case-insensitive
        self.assertEqual(keyboard_layout.get("1"), 2)
        self.assertEqual(keyboard_layout.get("KeY_1"), 2)

        self.assertEqual(keyboard_layout.get("AlT_L"), 56)
        self.assertEqual(keyboard_layout.get("KEy_LEFtALT"), 56)

        self.assertEqual(keyboard_layout.get("kEY_LeFTSHIFT"), 42)
        self.assertEqual(keyboard_layout.get("ShiFt_L"), 42)

        self.assertEqual(keyboard_layout.get("BTN_left"), 272)

        self.assertIsNotNone(keyboard_layout.get("KEY_KP4"))
        self.assertEqual(keyboard_layout.get("KP_Left"), keyboard_layout.get("KEY_KP4"))

        # this only lists the correct casing,
        # includes linux constants and xmodmap symbols
        names = keyboard_layout.list_names()
        self.assertIn("2", names)
        self.assertIn("c", names)
        self.assertIn("KEY_3", names)
        self.assertNotIn("key_3", names)
        self.assertIn("KP_Down", names)
        self.assertNotIn("kp_down", names)
        names = keyboard_layout._mapping.keys()
        self.assertIn("F4", names)
        self.assertNotIn("f4", names)
        self.assertIn("BTN_RIGHT", names)
        self.assertNotIn("btn_right", names)
        self.assertIn("KEY_KP7", names)
        self.assertIn("KP_Home", names)
        self.assertNotIn("kp_home", names)

        self.assertEqual(keyboard_layout.get("disable"), -1)

    def test_get_name_no_xmodmap(self):
        # if xmodmap is not installed, uses the linux constant names
        keyboard_layout = KeyboardLayout()

        def check_output(*args, **kwargs):
            raise FileNotFoundError

        with patch.object(subprocess, "check_output", check_output):
            keyboard_layout.populate()
            self.assertEqual(keyboard_layout.get_name(KEY_A), "KEY_A")

            # testing for BTN_LEFT is especially important, because
            # `evdev.ecodes.BTN.get(code)` returns an array of ['BTN_LEFT', 'BTN_MOUSE']
            self.assertEqual(keyboard_layout.get_name(BTN_LEFT), "BTN_LEFT")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
from typing import Dict, List

from inputremapper.groups import DeviceType

# useful type aliases
Capabilities = Dict[int, List]
Name = str
Key = str
DeviceTypes = List[DeviceType]


class MessageType(Enum):
    reset_gui = "reset_gui"
    terminate = "terminate"
    init = "init"

    uinputs = "uinputs"
    groups = "groups"
    group = "group"
    preset = "preset"
    mapping = "mapping"
    selected_event = "selected_event"
    combination_recorded = "combination_recorded"

    # only the reader_client should send those messages:
    recording_started = "recording_started"
    recording_finished = "recording_finished"

    combination_update = "combination_update"
    status_msg = "status_msg"
    injector_state = "injector_state"

    gui_focus_request = "gui_focus_request"
    user_confirm_request = "user_confirm_request"

    do_stack_switch = "do_stack_switch"

    # for unit tests:
    test1 = "test1"
    test2 = "test2"



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple

import evdev
from evdev.ecodes import EV_ABS

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.utils import get_evdev_constant_name


class AbsToBtnHandler(MappingHandler):
    """Handler which transforms an EV_ABS to a button event."""

    _input_config: InputConfig
    _active: bool
    _sub_handler: InputEventHandler

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ):
        super().__init__(combination, mapping, global_uinputs)

        self._active = False
        self._input_config = combination[0]
        assert self._input_config.analog_threshold
        assert len(combination) == 1

    def __str__(self):
        name = get_evdev_constant_name(*self._input_config.type_and_code)
        return f'AbsToBtnHandler for "{name}" ' f"{self._input_config.type_and_code}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return self._sub_handler

    def _trigger_point(self, abs_min: int, abs_max: int) -> Tuple[float, float]:
        """Calculate the axis mid and trigger point."""
        #  TODO: potentially cache this function
        assert self._input_config.analog_threshold
        if abs_min == -1 and abs_max == 1:
            # this is a hat switch
            # return +-1
            return (
                self._input_config.analog_threshold
                // abs(self._input_config.analog_threshold),
                0,
            )

        half_range = (abs_max - abs_min) / 2
        middle = half_range + abs_min
        trigger_offset = half_range * self._input_config.analog_threshold / 100

        # threshold, middle
        return middle + trigger_offset, middle

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if event.input_match_hash != self._input_config.input_match_hash:
            return False

        absinfo = {
            entry[0]: entry[1] for entry in source.capabilities(absinfo=True)[EV_ABS]
        }
        threshold, mid_point = self._trigger_point(
            absinfo[event.code].min, absinfo[event.code].max
        )
        value = event.value
        if (value < threshold > mid_point) or (value > threshold < mid_point):
            if self._active:
                event = event.modify(value=0, actions=(EventActions.as_key,))
            else:
                # consume the event.
                # We could return False to forward events
                return True
        else:
            if value >= threshold > mid_point:
                direction = EventActions.positive_trigger
            else:
                direction = EventActions.negative_trigger
            event = event.modify(value=1, actions=(EventActions.as_key, direction))

        self._active = bool(event.value)
        # logger.debug(event.event_tuple, "sending to sub_handler")
        return self._sub_handler.notify(
            event,
            source=source,
            suppress=suppress,
        )

    def reset(self) -> None:
        self._active = False
        self._sub_handler.reset()
